"""
activeslicehandle.py
Created by Shawn on 2011-02-05.
"""

# from exceptions import IndexError
from math import floor

from cadnano2.controllers.itemcontrollers.activesliceitemcontroller import ActiveSliceItemController
from cadnano2.views import styles
import cadnano2.util as util

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt', 'QObject',
                                        'pyqtSignal', 'pyqtSlot', 'QEvent'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QFont',
                                       'QPen',
                                       'QDrag',
                                       'QUndoCommand'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsSimpleTextItem',
                                           'QGraphicsRectItem'])


_baseWidth = styles.PATH_BASE_WIDTH
_brush = QBrush(styles.activeslicehandlefill)
_labelbrush = QBrush(styles.orangestroke)
_pen = QPen(styles.activeslicehandlestroke,\
            styles.SLICE_HANDLE_STROKE_WIDTH)
_font = QFont(styles.thefont, 12, QFont.Weight.Bold)


class ActiveSliceItem(QGraphicsRectItem):
    """ActiveSliceItem for the Path View"""

    def __init__(self, partItem, activeBaseIndex):
        super(ActiveSliceItem, self).__init__(partItem)
        self._partItem = partItem
        self._activeTool = partItem.activeTool()
        self._activeSlice = 0
        self._lowDragBound = 0
        self._highDragBound = self.part().maxBaseIdx()
        self._controller = ActiveSliceItemController(self, partItem.part())

        self._label = QGraphicsSimpleTextItem("", parent=self)
        self._label.setPos(0, -18)
        self._label.setFont(_font)
        self._label.setBrush(_labelbrush)
        self._label.hide()

        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setAcceptHoverEvents(True)
        self.setZValue(styles.ZACTIVESLICEHANDLE)
        self.setRect(QRectF(0, 0, _baseWidth,\
                      self._partItem.boundingRect().height()))
        self.setPos(activeBaseIndex*_baseWidth, 0)
        self.setBrush(_brush)
        self.setPen(_pen)

        # reuse select tool methods for other tools
        self.addSeqToolMousePress = self.selectToolMousePress
        self.addSeqToolMouseMove = self.selectToolMouseMove
        self.breakToolMousePress = self.selectToolMousePress
        self.breakToolMouseMove = self.selectToolMouseMove
        self.insertionToolMousePress = self.selectToolMousePress
        self.insertionToolMouseMove = self.selectToolMouseMove
        self.paintToolMousePress = self.selectToolMousePress
        self.paintToolMouseMove = self.selectToolMouseMove
        self.pencilToolMousePress = self.selectToolMousePress
        self.pencilToolMouseMove = self.selectToolMouseMove
        self.skipToolMousePress = self.selectToolMousePress
        self.skipToolMouseMove = self.selectToolMouseMove
    # end def

    ### SLOTS ###
    def strandChangedSlot(self, sender, vh):
        pass
    # end def

    def updateRectSlot(self, part):
        bw = _baseWidth
        newRect = QRectF(0, 0, bw,\
                    self._partItem.virtualHelixBoundingRect().height())
        if newRect != self.rect():
            self.setRect(newRect)
        self._hideIfEmptySelection()
        self.updateIndexSlot(part, part.activeBaseIndex())
        return newRect
    # end def

    def updateIndexSlot(self, part, baseIndex):
        """The slot that receives active slice changed notifications from
        the part and changes the receiver to reflect the part"""
        label = self._label
        bw = _baseWidth
        bi = util.clamp(int(baseIndex), 0, self.part().maxBaseIdx())
        self.setPos(bi * bw, -styles.PATH_HELIX_PADDING)
        self._activeSlice = bi
        if label:
            label.setText("%d" % bi)
            label.setX((bw - label.boundingRect().width()) / 2)
    # end def

    ### ACCESSORS ###
    def activeBaseIndex(self):
        return self.part().activeBaseIndex()
    # end def

    def part(self):
        return self._partItem.part()
    # end def

    def partItem(self):
        return self._partItem
    # end def

    ### PUBLIC METHODS FOR DRAWING / LAYOUT ###
    def removed(self):
        scene = self.scene()
        scene.removeItem(self._label)
        scene.removeItem(self)
        self._partItem = None
        self._label = None
        self._controller.disconnectSignals()
        self.controller = None
    # end def

    def resetBounds(self):
        """Call after resizing virtualhelix canvas."""
        self._highDragBound = self.part().maxBaseIdx()
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _hideIfEmptySelection(self):
        vis = self.part().numberOfVirtualHelices() > 0
        self.setVisible(vis)
        self._label.setVisible(vis)
    # end def

    def _setActiveBaseIndex(self, baseIndex):
        self.part().setActiveBaseIndex(baseIndex)
    # end def

    ### EVENT HANDLERS ###
    def hoverEnterEvent(self, event):
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        self._partItem.updateStatusBar("%d" % self.part().activeBaseIndex())
        QGraphicsItem.hoverEnterEvent(self, event)
    # end def

    def hoverLeaveEvent(self, event):
        self.setCursor(Qt.CursorShape.ArrowCursor)
        self._partItem.updateStatusBar("")
        QGraphicsItem.hoverLeaveEvent(self, event)
    # end def

    def mousePressEvent(self, event):
        """
        Parses a mousePressEvent, calling the approproate tool method as
        necessary. Stores _moveIdx for future comparison.
        """
        if event.button() != Qt.MouseButton.LeftButton:
            event.ignore()
            QGraphicsItem.mousePressEvent(self, event)
            return
        self.scene().views()[0].addToPressList(self)
        self._moveIdx = int(floor((self.x()+event.pos().x()) / _baseWidth))
        toolMethodName = str(self._activeTool()) + "MousePress"
        if hasattr(self, toolMethodName):  # if the tool method exists
            modifiers = event.modifiers()
            getattr(self, toolMethodName)(modifiers)  # call tool method

    def mouseMoveEvent(self, event):
        """
        Parses a mouseMoveEvent, calling the approproate tool method as
        necessary. Updates _moveIdx if it changed.
        """
        toolMethodName = str(self._activeTool()) + "MouseMove"
        if hasattr(self, toolMethodName):  # if the tool method exists
            idx = int(floor((self.x()+event.pos().x()) / _baseWidth))
            if idx != self._moveIdx:  # did we actually move?
                modifiers = event.modifiers()
                self._moveIdx = idx
                getattr(self, toolMethodName)(modifiers, idx)  # call tool method

    def customMouseRelease(self, event):
        """
        Parses a mouseReleaseEvent, calling the approproate tool method as
        necessary. Deletes _moveIdx if necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseRelease"
        if hasattr(self, toolMethodName):  # if the tool method exists
            modifiers = event.modifiers()
            x = event.position().x()
            getattr(self, toolMethodName)(modifiers, x)  # call tool method
        if self._moveIdx is not None:
            self._moveIdx = None

    ### TOOL METHODS ###
    def selectToolMousePress(self, modifiers):
        """
        Set the allowed drag bounds for use by selectToolMouseMove.
        """
        if (modifiers & Qt.KeyboardModifier.AltModifier) and (modifiers & Qt.KeyboardModifier.ShiftModifier):
            self.part().undoStack().beginMacro("Auto-drag Scaffold(s)")
            for vh in self.part().getVirtualHelices():
                # SCAFFOLD
                # resize 3' first
                for strand in vh.scaffoldStrandSet():
                    idx5p = strand.idx5Prime()
                    idx3p = strand.idx3Prime()
                    if not strand.hasXoverAt(idx3p):
                        lo, hi = strand.getResizeBounds(idx3p)
                        if strand.isDrawn5to3():
                            strand.resize((idx5p, hi))
                        else:
                            strand.resize((lo, idx5p))
                # resize 5' second
                for strand in vh.scaffoldStrandSet():
                    idx5p = strand.idx5Prime()
                    idx3p = strand.idx3Prime()
                    if not strand.hasXoverAt(idx5p):
                        lo, hi = strand.getResizeBounds(idx5p)
                        if strand.isDrawn5to3():
                            strand.resize((lo, idx3p))
                        else:
                            strand.resize((idx3p, hi))
                # STAPLE
                # resize 3' first
                for strand in vh.stapleStrandSet():
                    idx5p = strand.idx5Prime()
                    idx3p = strand.idx3Prime()
                    if not strand.hasXoverAt(idx3p):
                        lo, hi = strand.getResizeBounds(idx3p)
                        if strand.isDrawn5to3():
                            strand.resize((idx5p, hi))
                        else:
                            strand.resize((lo, idx5p))
                # resize 5' second
                for strand in vh.stapleStrandSet():
                    idx5p = strand.idx5Prime()
                    idx3p = strand.idx3Prime()
                    if not strand.hasXoverAt(idx3p):
                        lo, hi = strand.getResizeBounds(idx5p)
                        if strand.isDrawn5to3():
                            strand.resize((lo, idx3p))
                        else:
                            strand.resize((idx3p, hi))

            self.part().undoStack().endMacro()
    # end def

    def selectToolMouseMove(self, modifiers, idx):
        """
        Given a new index (pre-validated as different from the prev index),
        calculate the new x coordinate for self, move there, and notify the
        parent strandItem to redraw its horizontal line.
        """
        idx = util.clamp(idx, self._lowDragBound, self._highDragBound)
        x = int(idx * _baseWidth)
        self.setPos(x, self.y())
        self.updateIndexSlot(None, idx)
        self._setActiveBaseIndex(idx)
        self._partItem.updateStatusBar("%d" % self.part().activeBaseIndex())
    # end def



class ViewRootController():
    def __init__(self, viewRoot, modelDocument):
        self._viewRoot = viewRoot
        self._modelDocument = modelDocument
        self.connectSignals()

    def connectSignals(self):
        mD = self._modelDocument
        vR = self._viewRoot
        mD.documentPartAddedSignal.connect(vR.partAddedSlot)
        mD.documentClearSelectionsSignal.connect(vR.clearSelectionsSlot)
        mD.documentSelectionFilterChangedSignal.connect(vR.selectionFilterChangedSlot)
        mD.documentViewResetSignal.connect(vR.resetRootItemSlot)

    def disconnectSignals(self):
        mD = self._modelDocument
        vR = self._viewRoot
        mD.documentPartAddedSignal.disconnect(vR.partAddedSlot)
        mD.documentClearSelectionsSignal.disconnect(vR.clearSelectionsSlot)
        mD.documentSelectionFilterChangedSignal.disconnect(vR.selectionFilterChangedSlot)
        mD.documentViewResetSignal.disconnect(vR.resetRootItemSlot)


#!/usr/bin/env python
# encoding: utf-8

from operator import attrgetter
import cadnano2.util as util
from array import array
from .decorators.insertion import Insertion

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QUndoStack', 'QUndoCommand'])


ARRAY_TYPE = 'B'


def sixb(x): return x.encode('utf-8')


def tostring(x): return x.tobytes().decode('utf-8')


class Strand(QObject):
    """
    A Strand is a continuous stretch of bases that are all in the same
    StrandSet (recall: a VirtualHelix is made up of two StrandSets).

    Every Strand has two endpoints. The naming convention for keeping track
    of these endpoints is based on the relative numeric value of those
    endpoints (low and high). Thus, Strand has a '_baseIdxLow', which is its
    index with the lower numeric value (typically positioned on the left),
    and a '_baseIdxHigh' which is the higher-value index (typically positioned
    on the right)

    Strands can be linked to other strands by "connections". References to
    connected strands are named "_strand5p" and "_strand3p", which correspond
    to the 5' and 3' phosphate linkages in the physical DNA strand,
    respectively. Since Strands can point 5'-to-3' in either the low-to-high
    or high-to-low directions, connection accessor methods (connectionLow and
    connectionHigh) are bound during the init for convenience.
    """

    def __init__(self, strandSet, baseIdxLow, baseIdxHigh, oligo=None):
        super(Strand, self).__init__(strandSet)
        self._strandSet = strandSet
        self._doc = strandSet.document()

        self._baseIdxLow = baseIdxLow  # base index of the strand's left bound
        self._baseIdxHigh = baseIdxHigh  # base index of the right bound
        self._oligo = oligo
        self._strand5p = None  # 5' connection to another strand
        self._strand3p = None  # 3' connection to another strand
        self._sequence = None

        self._decorators = {}
        self._modifiers = {}

        # dynamic methods for mapping high/low connection /indices
        # to corresponding 3Prime 5Prime
        isDrawn5to3 = strandSet.isDrawn5to3()
        if isDrawn5to3:
            self.idx5Prime = self.lowIdx
            self.idx3Prime = self.highIdx
            self.connectionLow = self.connection5p
            self.connectionHigh = self.connection3p
            self.setConnectionLow = self.setConnection5p
            self.setConnectionHigh = self.setConnection3p
        else:
            self.idx5Prime = self.highIdx
            self.idx3Prime = self.lowIdx
            self.connectionLow = self.connection3p
            self.connectionHigh = self.connection5p
            self.setConnectionLow = self.setConnection3p
            self.setConnectionHigh = self.setConnection5p
        self._isDrawn5to3 = isDrawn5to3
    # end def

    def __repr__(self):
        clsName = self.__class__.__name__
        s = "%s.<%s(%s, %s)>" % (self._strandSet.__repr__(),
                                clsName,
                                self._baseIdxLow,
                                self._baseIdxHigh)
        return s

    def generator5pStrand(self):
        """
        Iterate from self to the final _strand5p == None
        3prime to 5prime
        Includes originalCount to check for circular linked list
        """
        originalCount = 0
        node0 = node = self
        f = attrgetter('_strand5p')
        # while node and originalCount == 0:
        #     yield node  # equivalent to: node = node._strand5p
        #     node = f(node)
        #     if node0 == self:
        #         originalCount += 1
        while node:
            yield node  # equivalent to: node = node._strand5p
            node = f(node)
            if node0 == node:
                break

    # end def

    def generator3pStrand(self):
        """
        Iterate from self to the final _strand3p == None
        5prime to 3prime
        Includes originalCount to check for circular linked list
        """
        originalCount = 0
        node0 = node = self
        f = attrgetter('_strand3p')
        while node:
            yield node  # equivalent to: node = node._strand5p
            node = f(node)
            if node0 == node:
                break
    # end def

    def strandFilter(self):
        return self._strandSet.strandFilter()

    ### SIGNALS ###
    strandHasNewOligoSignal = pyqtSignal(QObject)  # strand
    strandRemovedSignal = pyqtSignal(QObject)  # strand
    strandResizedSignal = pyqtSignal(QObject, tuple)

    # Parameters: (strand3p, strand5p)
    strandXover5pChangedSignal = pyqtSignal(QObject, QObject)
    strandXover5pRemovedSignal = pyqtSignal(QObject, QObject)

    # Parameters: (strand)
    strandUpdateSignal = pyqtSignal(QObject)

    # Parameters: (strand, insertion object)
    strandInsertionAddedSignal = pyqtSignal(QObject, object)
    strandInsertionChangedSignal = pyqtSignal(QObject, object)
    # Parameters: (strand, insertion index)
    strandInsertionRemovedSignal = pyqtSignal(QObject, int)

    # Parameters: (strand, decorator object)
    strandDecoratorAddedSignal = pyqtSignal(QObject, object)
    strandDecoratorChangedSignal = pyqtSignal(QObject, object)
    # Parameters: (strand, decorator index)
    strandDecoratorRemovedSignal = pyqtSignal(QObject, int)

    # Parameters: (strand, modifier object)
    strandModifierAddedSignal = pyqtSignal(QObject, object)
    strandModifierChangedSignal = pyqtSignal(QObject, object)
    # Parameters: (strand, modifier index)
    strandModifierRemovedSignal = pyqtSignal(QObject, int)

    # Parameters: (strand, value)
    selectedChangedSignal = pyqtSignal(QObject, tuple)

    ### SLOTS ###
    ### ACCESSORS ###
    def undoStack(self):
        return self._strandSet.undoStack()

    def decorators(self):
        return self._decorators
    # end def

    def isStaple(self):
        return self._strandSet.isStaple()

    def isScaffold(self):
        return self._strandSet.isScaffold()

    def part(self):
        return self._strandSet.part()
    # end def

    def document(self):
        return self._doc
    # end def

    def oligo(self):
        return self._oligo
    # end def

    def sequence(self, forExport=False):
        seq = self._sequence
        if seq:
            return util.markwhite(seq) if forExport else seq
        elif forExport:
            return ''.join(['?' for x in range(self.totalLength())])
        return ''
    # end def

    def strandSet(self):
        return self._strandSet
    # end def

    def strandType(self):
        return self._strandSet.strandType()

    def virtualHelix(self):
        return self._strandSet.virtualHelix()
    # end def

    def setSequence(self, sequenceString):
        """
        Applies sequence string from 5' to 3'
        return the tuple (used, unused) portion of the sequenceString
        """
        if sequenceString == None:
            self._sequence = None
            return None, None
        length = self.totalLength()
        if len(sequenceString) < length:
            bonus = length - len(sequenceString)
            sequenceString += ''.join([' ' for x in range(bonus)])
        temp = sequenceString[0:length]
        self._sequence = temp
        return temp, sequenceString[length:]
    # end def

    def reapplySequence(self):
        """
        """
        compSS = self.strandSet().complementStrandSet()

        # the strand sequence will need to be regenerated from scratch
        # as there are no guarantees about the entirety of the strand moving
        # i.e. both endpoints thanks to multiple selections so just redo the 
        # whole thing
        self._sequence = None

        for compStrand in compSS._findOverlappingRanges(self):
            compSeq = compStrand.sequence()
            usedSeq = util.comp(compSeq) if compSeq else None
            usedSeq = self.setComplementSequence(
                                        usedSeq, compStrand)
        # end for
    # end def

    def getComplementStrands(self):
        """
        return the list of complement strands that overlap with this strand
        """
        compSS = self.strandSet().complementStrandSet()
        return [compStrand for compStrand in compSS._findOverlappingRanges(self)]
    # end def

    def getPreDecoratorIdxList(self):
        """
        Return positions where predecorators should be displayed. This is
        just a very simple check for the presence of xovers on the strand.

        Will refine later by checking for lattice neighbors in 3D.
        """
        return list(range(self._baseIdxLow, self._baseIdxHigh + 1))
    # end def

    # def getPreDecoratorIdxList(self):
    #     """Return positions where predecorators should be displayed."""
    #     part = self._strandSet.part()
    #     validIdxs = sorted([idx[0] for idx in part._stapL + part._stapH])
    #     lo, hi = self._baseIdxLow, self._baseIdxHigh
    #     start = lo if self.connectionLow() == None else lo+1
    #     end = hi if self.connectionHigh() == None else hi-1
    #     ret = []
    #     for i in range(start, end+1):
    #         if i % part.stepSize() in validIdxs:
    #             ret.append(i)
    #     return ret
    # # end def

    def setComplementSequence(self, sequenceString, strand):
        """
        This version takes anothers strand and only sets the indices that
        align with the given complimentary strand

        return the used portion of the sequenceString

        As it depends which direction this is going, and strings are stored in
        memory left to right, we need to test for isDrawn5to3 to map the
        reverse compliment appropriately, as we traverse overlapping strands.

        We reverse the sequence ahead of time if we are applying it 5' to 3',
        otherwise we reverse the sequence post parsing if it's 3' to 5'

        Again, sequences are stored as strings in memory 5' to 3' so we need
        to jump through these hoops to iterate 5' to 3' through them correctly

        Perhaps it's wiser to merely store them left to right and reverse them
        at draw time, or export time
        """
        sLowIdx, sHighIdx = self._baseIdxLow, self._baseIdxHigh
        cLowIdx, cHighIdx = strand.idxs()

        # get the ovelap
        lowIdx, highIdx = util.overlap(sLowIdx, sHighIdx, cLowIdx, cHighIdx)

        # only get the characters we're using, while we're at it, make it the
        # reverse compliment

        totalLength = self.totalLength()

        # see if we are applying
        if sequenceString is None:
            # clear out string for in case of not total overlap
            useSeq = ''.join([' ' for x in range(totalLength)])
        else:  # use the string as is
            useSeq = sequenceString[::-1] if self._isDrawn5to3 \
                                            else sequenceString

        temp = array(ARRAY_TYPE, sixb(useSeq))
        if self._sequence is None:
            tempSelf = array(ARRAY_TYPE, sixb(''.join([' ' for x in range(totalLength)])))
        else:
            tempSelf = array(ARRAY_TYPE, sixb(self._sequence) if self._isDrawn5to3 \
                                                    else sixb(self._sequence[::-1]))

        # generate the index into the compliment string
        a = self.insertionLengthBetweenIdxs(sLowIdx, lowIdx - 1)
        b = self.insertionLengthBetweenIdxs(lowIdx, highIdx)
        c = strand.insertionLengthBetweenIdxs(cLowIdx, lowIdx - 1)
        start = lowIdx - cLowIdx + c
        end = start + b + highIdx - lowIdx + 1
        tempSelf[lowIdx - sLowIdx + a:highIdx - sLowIdx + 1 + a + b] = temp[start:end]
        # print "old sequence", self._sequence
        self._sequence = tostring(tempSelf)

        # if we need to reverse it do it now
        if not self._isDrawn5to3:
            self._sequence = self._sequence[::-1]

        # test to see if the string is empty(), annoyingly expensive
        if len(self._sequence.strip()) == 0:
            self._sequence = None

        # print "new sequence", self._sequence
        return self._sequence
    # end def

    ### PUBLIC METHODS FOR QUERYING THE MODEL ###
    def connection3p(self):
        return self._strand3p
    # end def

    def connection5p(self):
        return self._strand5p
    # end def

    def idxs(self):
        return (self._baseIdxLow, self._baseIdxHigh)
    # end def

    def lowIdx(self):
        return self._baseIdxLow
    # end def

    def highIdx(self):
        return self._baseIdxHigh
    # end def

    def idx3Prime(self):
        """Returns the absolute baseIdx of the 3' end of the strand."""
        return self.idx3Prime

    def idx5Prime(self):
        """Returns the absolute baseIdx of the 5' end of the strand."""
        return self.idx5Prime

    def isDrawn5to3(self):
        return self._strandSet.isDrawn5to3()
    # end def

    def getSequenceList(self):
        """
        return the list of sequences strings comprising the sequence and the
        inserts as a tuple with the index of the insertion
        [(idx, (strandItemString, insertionItemString), ...]

        This takes advantage of the fact the python iterates a dictionary
        by keys in order so if keys are indices, the insertions will iterate
        out from low index to high index
        """
        seqList = []
        isDrawn5to3 = self._isDrawn5to3
        seq = self._sequence if isDrawn5to3 else self._sequence[::-1]
        # assumes a sequence has been applied correctly and is up to date
        tL = self.totalLength()

        offsetLast = 0
        lengthSoFar = 0
        iLength = 0
        lI, hI = self.idxs()

        for insertion in self.insertionsOnStrand():
            iLength = insertion.length()
            index = insertion.idx()
            offset = index + 1 - lI + lengthSoFar
            if iLength < 0:
                offset -= 1
            # end if
            lengthSoFar += iLength
            seqItem = seq[offsetLast:offset]  # the stranditem seq

            # Because skips literally skip displaying a character at a base
            # position, this needs to be accounted for seperately
            if iLength < 0:
                seqItem += ' '
                offsetLast = offset
            else:
                offsetLast = offset + iLength
            seqInsertion = seq[offset:offsetLast]  # the insertions sequence
            seqList.append((index, (seqItem, seqInsertion)))
        # end for
        # append the last bit of the strand
        seqList.append((lI + tL, (seq[offsetLast:tL], '')))
        if not isDrawn5to3:
            # reverse it again so all sub sequences are from 5' to 3'
            for i in range(len(seqList)):
                index, temp = seqList[i]
                seqList[i] = (index, (temp[0][::-1], temp[1][::-1]))
        return seqList
    # end def

    def canResizeTo(self, newLow, newHigh):
        """
        Checks to see if a resize is allowed. Similar to getResizeBounds
        but works for two bounds at once.
        """
        lowNeighbor, highNeighbor = self._strandSet.getNeighbors(self)
        lowBound = lowNeighbor.highIdx() if lowNeighbor \
                                            else self.part().minBaseIdx()
        highBound = highNeighbor.lowIdx() if highNeighbor \
                                            else self.part().maxBaseIdx()

        if newLow > lowBound and newHigh < highBound:
            return True
        return False

    def getResizeBounds(self, idx):
        """
        Determines (inclusive) low and high drag boundaries resizing
        from an endpoint located at idx.

        When resizing from _baseIdxLow:
            low bound is determined by checking for lower neighbor strands.
            high bound is the index of this strand's high cap, minus 1.

        When resizing from _baseIdxHigh:
            low bound is the index of this strand's low cap, plus 1.
            high bound is determined by checking for higher neighbor strands.

        When a neighbor is not present, just use the Part boundary.
        """
        neighbors = self._strandSet.getNeighbors(self)
        if idx == self._baseIdxLow:
            if neighbors[0]:
                low = neighbors[0].highIdx() + 1
            else:
                low = self.part().minBaseIdx()
            return low, self._baseIdxHigh - 1
        else:  # self._baseIdxHigh
            if neighbors[1]:
                high = neighbors[1].lowIdx() - 1
            else:
                high = self.part().maxBaseIdx()
            return self._baseIdxLow + 1, high
    # end def

    def hasXoverAt(self, idx):
        """
        An xover is necessarily at an enpoint of a strand
        """
        if idx == self.highIdx():
            return True if self.connectionHigh() != None else False
        elif idx == self.lowIdx():
            return True if self.connectionLow() != None else False
        else:
            return False
    # end def

    def canInstallXoverAt(self, idx, fromStrand, fromIdx):
        """
        Assumes idx is:
        self.lowIdx() <= idx <= self.highIdx()
        """

        if self.hasXoverAt(idx):
            return False
        sS = self.strandSet()
        isSameStrand = fromStrand == self
        isStrandTypeMatch = \
                fromStrand.strandSet().strandType() == sS.strandType() \
                                                if fromStrand else True
        if not isStrandTypeMatch:
            return False
        isDrawn5to3 = sS.isDrawn5to3()
        indexDiffH = self.highIdx() - idx
        indexDiffL = idx - self.lowIdx()
        index3Lim = self.idx3Prime() - 1 if isDrawn5to3 \
                                            else self.idx3Prime() + 1
        if isSameStrand:
            indexDiffStrands = fromIdx - idx
            if idx == self.idx5Prime() or idx == index3Lim:
                return True
            elif indexDiffStrands > -3 and indexDiffStrands < 3:
                return False
        # end if for same Strand
        if idx == self.idx5Prime() or idx == index3Lim:
            return True
        elif indexDiffH > 2 and indexDiffL > 1:
            return True
        else:
            return False
    #end def

    def insertionLengthBetweenIdxs(self, idxL, idxH):
        """
        includes the length of insertions in addition to the bases
        """
        tL = 0
        insertions = self.insertionsOnStrand(idxL, idxH)
        for insertion in insertions:
            tL += insertion.length()
        return tL
    # end def

    def insertionsOnStrand(self, idxL=None, idxH=None):
        """
        if passed indices it will use those as a bounds
        """
        insertions = []
        coord = self.virtualHelix().coord()
        insertionsDict = self.part().insertions()[coord]
        sortedIndices = sorted(insertionsDict.keys())
        if idxL == None:
            idxL, idxH = self.idxs()
        for index in sortedIndices:
            insertion = insertionsDict[index]
            if idxL <= insertion.idx() <= idxH:
                insertions.append(insertion)
            # end if
        # end for
        return insertions
    # end def

    def length(self):
        return self._baseIdxHigh - self._baseIdxLow + 1
    # end def

    def totalLength(self):
        """
        includes the length of insertions in addition to the bases
        """
        tL = 0
        insertions = self.insertionsOnStrand()

        for insertion in insertions:
            tL += insertion.length()
        return tL + self.length()
    # end def

    ### PUBLIC METHODS FOR EDITING THE MODEL ###
    def addDecorators(self, additionalDecorators):
        """Used to add decorators during a merge operation."""
        self._decorators.update(additionalDecorators)
    # end def

    def addInsertion(self, idx, length, useUndoStack=True):
        """
        Adds an insertion or skip at idx.
        length should be
            >0 for an insertion
            -1 for a skip
        """
        cmds = []
        idxLow, idxHigh = self.idxs()
        if idxLow <= idx <= idxHigh:
            if not self.hasInsertionAt(idx):
                # make sure length is -1 if a skip
                if length < 0:
                    length = -1
                if useUndoStack:
                    cmds.append(self.oligo().applySequenceCMD(None, useUndoStack=useUndoStack))
                    for strand in self.getComplementStrands():
                        cmds.append(strand.oligo().applySequenceCMD(None, useUndoStack=useUndoStack))
                cmds.append(Strand.AddInsertionCommand(self, idx, length))
                util.execCommandList(self, cmds, desc="Add Insertion",
                                     useUndoStack=useUndoStack)
            # end if
        # end if
    # end def

    def changeInsertion(self, idx, length, useUndoStack=True):
        cmds = []
        idxLow, idxHigh = self.idxs()
        if idxLow <= idx <= idxHigh:
            if self.hasInsertionAt(idx):
                if length == 0:
                    self.removeInsertion(idx)
                else:
                    # make sure length is -1 if a skip
                    if length < 0:
                        length = -1
                    cmds.append(self.oligo().applySequenceCMD(None, useUndoStack=useUndoStack))
                    for strand in self.getComplementStrands():
                        cmds.append(strand.oligo().applySequenceCMD(None, useUndoStack=useUndoStack))
                    cmds.append(
                            Strand.ChangeInsertionCommand(self, idx, length))
                    util.execCommandList(
                                        self, cmds, desc="Change Insertion",
                                        useUndoStack=useUndoStack)
            # end if
        # end if
    # end def
    
    def removeInsertion(self,  idx, useUndoStack=True):
        cmds = []
        idxLow, idxHigh = self.idxs()
        if idxLow <= idx <= idxHigh:
            if self.hasInsertionAt(idx):
                cmds.append(self.oligo().applySequenceCMD(None, useUndoStack=useUndoStack))
                for strand in self.getComplementStrands():
                    cmds.append(strand.oligo().applySequenceCMD(None, useUndoStack=useUndoStack))
                cmds.append(Strand.RemoveInsertionCommand(self, idx))
                util.execCommandList(
                                    self, cmds, desc="Remove Insertion",
                                    useUndoStack=useUndoStack)
            # end if
        # end if
    # end def

    def destroy(self):
        self.setParent(None)
        self.deleteLater()  # QObject also emits a destroyed() Signal
    # end def

    def merge(self, idx):
        """Check for neighbor, then merge if possible."""
        lowNeighbor, highNeighbor = self._strandSet.getNeighbors(self)
        # determine where to check for neighboring endpoint
        if idx == self._baseIdxLow:
            if lowNeighbor:
                if lowNeighbor.highIdx() == idx - 1:
                    self._strandSet.mergeStrands(self, lowNeighbor)
        elif idx == self._baseIdxHigh:
            if highNeighbor:
                if highNeighbor.lowIdx() == idx + 1:
                    self._strandSet.mergeStrands(self, highNeighbor)
        else:
            raise IndexError
    # end def

    def resize(self, newIdxs, useUndoStack=True):
        cmds = []
        if self.strandSet().isScaffold():
            cmds.append(self.oligo().applySequenceCMD(None))
        cmds += self.getRemoveInsertionCommands(newIdxs)
        cmds.append(Strand.ResizeCommand(self, newIdxs))
        util.execCommandList(
                            self, cmds, desc="Resize strand",
                            useUndoStack=useUndoStack)
    # end def

    def setConnection3p(self, strand):
        self._strand3p = strand
    # end def

    def setConnection5p(self, strand):
        self._strand5p = strand
    # end def

    def setIdxs(self, idxs):
        self._baseIdxLow = idxs[0]
        self._baseIdxHigh = idxs[1]
    # end def

    def setOligo(self, newOligo, emitSignal=True):
        self._oligo = newOligo
        if emitSignal:
            self.strandHasNewOligoSignal.emit(self)
    # end def

    def setStrandSet(self, strandSet):
        self._strandSet = strandSet
    # end def

    def split(self, idx, updateSequence=True):
        """Called by view items to split this strand at idx."""
        self._strandSet.splitStrand(self, idx, updateSequence)

    def updateIdxs(self, delta):
        self._baseIdxLow += delta
        self._baseIdxHigh += delta
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def getRemoveInsertionCommands(self, newIdxs):
        """
        Removes Insertions, Decorators, and Modifiers that have fallen out of
        range of newIdxs.

        For insertions, it finds the ones that have neither Staple nor Scaffold
        strands at the insertion idx as a result of the change of this
        strand to newIdxs

        """
        decs = self._decorators
        mods = self._modifiers
        cIdxL, cIdxH = self.idxs()
        nIdxL, nIdxH = newIdxs

        lowOut, highOut = False, False
        insertions = []
        if cIdxL < nIdxL < cIdxH:
            idxL, idxH = cIdxL, nIdxL - 1
            insertions += self.insertionsOnStrand(idxL, idxH)
        else:
            lowOut = True
        if cIdxL < nIdxH < cIdxH:
            idxL, idxH = nIdxH + 1, cIdxH
            insertions += self.insertionsOnStrand(idxL, idxH)
        else:
            highOut = True
        # this only called if both the above aren't true
        # if lowOut and highOut:
        # if we move the whole strand, just clear the insertions out
        if nIdxL > cIdxH or nIdxH < cIdxL:
            idxL, idxH = cIdxL, cIdxH
            insertions += self.insertionsOnStrand(idxL, idxH)
            # we stretched in this direction

        return self.clearInsertionsCommands(insertions, cIdxL, cIdxH)
    # end def

    def clearInsertionsCommands(self, insertions, idxL, idxH):
        """
        clear out insertions in this range
        """
        commands = []
        compSS = self.strandSet().complementStrandSet()

        overlappingStrandList = compSS.getOverlappingStrands(idxL, idxH)
        for insertion in insertions:
            idx = insertion.idx()
            removeMe = True
            for strand in overlappingStrandList:
                overLapIdxL, overLapIdxH = strand.idxs()
                if overLapIdxL <= idx <= overLapIdxH:
                    removeMe = False
                # end if
            # end for
            if removeMe:
                commands.append(Strand.RemoveInsertionCommand(self, idx))
            else:
                pass
                # print "keeping %s insertion at %d" % (self, key)
        # end for

        ### ADD CODE HERE TO HANDLE DECORATORS AND MODIFIERS
        return commands
    # end def

    def clearDecoratorCommands(self):
        insertions = self.insertionsOnStrand()
        return self.clearInsertionsCommands(insertions, *self.idxs())
    # end def

    def hasDecoratorAt(self, idx):
        return idx in self._decorators
    # end def

    def hasInsertion(self):
        """
        Iterate through dict of insertions for this strand's virtualhelix
        and return True of any of the indices overlap with the strand.
        """
        coord = self.virtualHelix().coord()
        insts = self.part().insertions()[coord]
        for i in range(self._baseIdxLow, self._baseIdxHigh+1):
            if i in insts:
                return True
        return False
    # end def

    def hasInsertionAt(self, idx):
        coord = self.virtualHelix().coord()
        insts = self.part().insertions()[coord]
        return idx in insts
    # end def

    def hasModifierAt(self, idx):
        return idx in self._modifiers
    # end def

    def shallowCopy(self):
        """
        can't use python module 'copy' as the dictionary _decorators
        needs to be shallow copied as well, but wouldn't be if copy.copy()
        is used, and copy.deepcopy is undesired
        """
        nS = Strand(self._strandSet, *self.idxs())
        nS._oligo = self._oligo
        nS._strand5p = self._strand5p
        nS._strand3p = self._strand3p
        # required to shallow copy the dictionary
        nS._decorators = dict(list(self._decorators.items()))
        nS._sequence = None  # self._sequence
        return nS
    # end def

    def deepCopy(self, strandSet, oligo):
        """
        can't use python module 'copy' as the dictionary _decorators
        needs to be shallow copied as well, but wouldn't be if copy.copy()
        is used, and copy.deepcopy is undesired
        """
        nS = Strand(strandSet, *self.idxs())
        nS._oligo = oligo
        decs = nS._decorators
        for key, decOrig in self._decorators:
            decs[key] = decOrig.deepCopy()
        # end fo
        nS._sequence = self._sequence
        return nS
    # end def

    ### COMMANDS ###
    class ResizeCommand(QUndoCommand):
        def __init__(self, strand, newIdxs):
            super(Strand.ResizeCommand, self).__init__()
            self.strand = strand
            self.oldIndices = oI = strand.idxs()
            self.newIdxs = newIdxs
            # an increase in length leads to positive delta
            self.delta = (newIdxs[1] - newIdxs[0]) - (oI[1] - oI[0])
            # now handle insertion deltas
            oldInsertions = strand.insertionsOnStrand(*oI)
            newInsertions = strand.insertionsOnStrand(*newIdxs)
            oL = 0
            for i in oldInsertions:
                oL += i.length()
            nL = 0
            for i in newInsertions:
                nL += i.length()
            self.delta += (nL-oL)

            # the strand sequence will need to be regenerated from scratch
            # as there are no guarantees about the entirety of the strand moving
            # thanks to multiple selections
        # end def

        def redo(self):
            std = self.strand
            nI = self.newIdxs
            strandSet = self.strand.strandSet()
            part = strandSet.part()

            std.oligo().incrementLength(self.delta)
            std.setIdxs(nI)
            if strandSet.isStaple():
                
                std.reapplySequence()
            std.strandResizedSignal.emit(std, nI)
            # for updating the Slice View displayed helices
            part.partStrandChangedSignal.emit(part, strandSet.virtualHelix())
            std5p = std.connection5p()
            if std5p:
                std5p.strandResizedSignal.emit(std5p, std5p.idxs())
        # end def

        def undo(self):
            std = self.strand
            oI = self.oldIndices
            strandSet = self.strand.strandSet()
            part = strandSet.part()

            std.oligo().decrementLength(self.delta)
            std.setIdxs(oI)
            if strandSet.isStaple():
                std.reapplySequence()
            std.strandResizedSignal.emit(std, oI)
            # for updating the Slice View displayed helices
            part.partStrandChangedSignal.emit(part, strandSet.virtualHelix())
            std5p = std.connection5p()
            if std5p:
                std5p.strandResizedSignal.emit(std5p, std5p.idxs())
        # end def
    # end class

    class AddInsertionCommand(QUndoCommand):
        def __init__(self, strand, idx, length):
            super(Strand.AddInsertionCommand, self).__init__()
            self._strand = strand
            coord = strand.virtualHelix().coord()
            self._insertions = strand.part().insertions()[coord]
            self._idx = idx
            self._length = length
            self._insertion = Insertion(idx, length)
            self._compStrand = \
                        strand.strandSet().complementStrandSet().getStrand(idx)
        # end def

        def redo(self):
            strand = self._strand
            cStrand = self._compStrand
            inst = self._insertion
            self._insertions[self._idx] = inst
            strand.oligo().incrementLength(inst.length())
            strand.strandInsertionAddedSignal.emit(strand, inst)
            if cStrand:
                cStrand.oligo().incrementLength(inst.length())
                cStrand.strandInsertionAddedSignal.emit(cStrand, inst)
        # end def

        def undo(self):
            strand = self._strand
            cStrand = self._compStrand
            inst = self._insertion
            strand.oligo().decrementLength(inst.length())
            if cStrand:
                cStrand.oligo().decrementLength(inst.length())
            idx = self._idx
            del self._insertions[idx]
            strand.strandInsertionRemovedSignal.emit(strand, idx)
            if cStrand:
                cStrand.strandInsertionRemovedSignal.emit(cStrand, idx)
        # end def
    # end class

    class RemoveInsertionCommand(QUndoCommand):
        def __init__(self, strand, idx):
            super(Strand.RemoveInsertionCommand, self).__init__()
            self._strand = strand
            self._idx = idx
            coord = strand.virtualHelix().coord()
            self._insertions = strand.part().insertions()[coord]
            self._insertion = self._insertions[idx]
            self._compStrand = \
                        strand.strandSet().complementStrandSet().getStrand(idx)
        # end def

        def redo(self):
            strand = self._strand
            cStrand = self._compStrand
            inst = self._insertion
            strand.oligo().decrementLength(inst.length())
            if cStrand:
                cStrand.oligo().decrementLength(inst.length())
            idx = self._idx
            del self._insertions[idx]
            strand.strandInsertionRemovedSignal.emit(strand, idx)
            if cStrand:
                cStrand.strandInsertionRemovedSignal.emit(cStrand, idx)
        # end def

        def undo(self):
            strand = self._strand
            cStrand = self._compStrand
            coord = strand.virtualHelix().coord()
            inst = self._insertion
            strand.oligo().incrementLength(inst.length())
            self._insertions[self._idx] = inst
            strand.strandInsertionAddedSignal.emit(strand, inst)
            if cStrand:
                cStrand.oligo().incrementLength(inst.length())
                cStrand.strandInsertionAddedSignal.emit(cStrand, inst)
        # end def
    # end class

    class ChangeInsertionCommand(QUndoCommand):
        """
        Changes the length of an insertion to a non-zero value
        the caller of this needs to handle the case where a zero length
        is required and call RemoveInsertionCommand
        """
        def __init__(self, strand, idx, newLength):
            super(Strand.ChangeInsertionCommand, self).__init__()
            self._strand = strand
            coord = strand.virtualHelix().coord()
            self._insertions = strand.part().insertions()[coord]
            self._idx = idx
            self._newLength = newLength
            self._oldLength = self._insertions[idx].length()
            self._compStrand = \
                        strand.strandSet().complementStrandSet().getStrand(idx)
        # end def

        def redo(self):
            strand = self._strand
            cStrand = self._compStrand
            inst = self._insertions[self._idx]
            inst.setLength(self._newLength)
            strand.oligo().incrementLength(self._newLength - self._oldLength)
            strand.strandInsertionChangedSignal.emit(strand, inst)
            if cStrand:
                cStrand.oligo().incrementLength(
                                            self._newLength - self._oldLength)
                cStrand.strandInsertionChangedSignal.emit(cStrand, inst)
        # end def

        def undo(self):
            strand = self._strand
            cStrand = self._compStrand
            inst = self._insertions[self._idx]
            inst.setLength(self._oldLength)
            strand.oligo().decrementLength(self._newLength - self._oldLength)
            strand.strandInsertionChangedSignal.emit(strand, inst)
            if cStrand:
                cStrand.oligo().decrementLength(
                                            self._newLength - self._oldLength)
                cStrand.strandInsertionChangedSignal.emit(cStrand, inst)
        # end def
    # end class
# end class



import random
from operator import itemgetter
from itertools import repeat

from .strand import Strand
from .oligo import Oligo
from .enum import StrandType
from cadnano2.views import styles

import cadnano2.util as util
# import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QUndoStack', 'QUndoCommand'])


class StrandSet(QObject):
    """
    StrandSet is a container class for Strands, and provides the several
    publicly accessible methods for editing strands, including operations
    for creation, destruction, resizing, splitting, and merging strands.

    Views may also query StrandSet for information that is useful in
    determining if edits can be made, such as the bounds of empty space in
    which a strand can be created or resized.
    """
    def __init__(self, strandType, virtualHelix):
        super(StrandSet, self).__init__(virtualHelix)
        self._virtualHelix = virtualHelix
        self._doc = virtualHelix.document()
        self._strandList = []
        self._undoStack = None
        self._lastStrandSetIndex = None
        self._strandType = strandType
    # end def

    def __iter__(self):
        """Iterate over each strand in the strands list."""
        return self._strandList.__iter__()
    # end def

    def __repr__(self):
        if self._strandType == 0:
            type = 'scaf'
        else:
            type = 'stap'
        num = self._virtualHelix.number()
        return "<%s_StrandSet(%d)>" % (type, num)
    # end def

    ### SIGNALS ###
    strandsetStrandAddedSignal = pyqtSignal(QObject, QObject)  # strandset, strand

    ### SLOTS ###

    ### ACCESSORS ###
    def part(self):
        return self._virtualHelix.part()
    # end def

    def document(self):
        return self._doc
    # end def

    def generatorStrand(self):
        """Return a generator that yields the strands in self._strandList."""
        return iter(self._strandList)
    # end def

    ### PUBLIC METHODS FOR QUERYING THE MODEL ###
    def isDrawn5to3(self):
        return self._virtualHelix.isDrawn5to3(self)
    # end def

    def isStaple(self):
        return self._strandType == StrandType.Staple
    # end def

    def isScaffold(self):
        return self._strandType == StrandType.Scaffold
    # end def

    def getNeighbors(self, strand):
        isInSet, overlap, strandSetIdx = self._findIndexOfRangeFor(strand)
        sList = self._strandList
        if isInSet:
            if strandSetIdx > 0:
                lowStrand = sList[strandSetIdx - 1]
            else:
                lowStrand = None
            if strandSetIdx < len(sList) - 1:
                highStrand = sList[strandSetIdx + 1]
            else:
                highStrand = None
            return lowStrand, highStrand
        else:
            raise IndexError
    # end def

    def complementStrandSet(self):
        """
        Returns the complementary strandset. Used for insertions and
        sequence application.
        """
        vh = self.virtualHelix()
        if self.isStaple():
            return vh.scaffoldStrandSet()
        else:
            return vh.stapleStrandSet()
    # end def

    def getBoundsOfEmptyRegionContaining(self, baseIdx):
        """
        Returns the (tight) bounds of the contiguous stretch of unpopulated
        bases that includes the baseIdx.
        """
        lowIdx, highIdx = 0, self.partMaxBaseIdx()  # init the return values
        lenStrands = len(self._strandList)

        # not sure how to set this up this to help in caching
        # lastIdx = self._lastStrandSetIndex

        if lenStrands == 0:  # empty strandset, just return the part bounds
            return (lowIdx, highIdx)

        low = 0              # index of the first (left-most) strand
        high = lenStrands    # index of the last (right-most) strand
        while low < high:    # perform binary search to find empty region
            mid = (low + high) // 2
            midStrand = self._strandList[mid]
            mLow, mHigh = midStrand.idxs()
            if baseIdx < mLow:  # baseIdx is to the left of crntStrand
                high = mid   # continue binary search to the left
                highIdx = mLow - 1  # set highIdx to left of crntStrand
            elif baseIdx > mHigh:   # baseIdx is to the right of crntStrand
                low = mid + 1    # continue binary search to the right
                lowIdx = mHigh + 1  # set lowIdx to the right of crntStrand
            else:
                return (None, None)  # baseIdx was not empty
        self._lastStrandSetIndex = (low + high) // 2  # set cache
        return (lowIdx, highIdx)
    # end def

    def indexOfRightmostNonemptyBase(self):
        """Returns the high baseIdx of the last strand, or 0."""
        if len(self._strandList) > 0:
            return self._strandList[-1].highIdx()
        else:
            return 0

    def partMaxBaseIdx(self):
        """Return the bounds of the StrandSet as defined in the part."""
        return self._virtualHelix.part().maxBaseIdx()
    # end def

    def strandCount(self):
        return len(self._strandList)
    # end def

    def strandType(self):
        return self._strandType
    # end def

    ### PUBLIC METHODS FOR EDITING THE MODEL ###
    def createStrand(self, baseIdxLow, baseIdxHigh, useUndoStack=True):
        """
        Assumes a strand is being created at a valid set of indices.
        """
        boundsLow, boundsHigh = \
                            self.getBoundsOfEmptyRegionContaining(baseIdxLow)
        canInsert, strandSetIdx = \
                                self.getIndexToInsert(baseIdxLow, baseIdxHigh)
        if canInsert:
            c = StrandSet.CreateStrandCommand(self,
                                        baseIdxLow, baseIdxHigh, strandSetIdx)
            row, col = self._virtualHelix.coord()
            # d = "(%d,%d).%d + [%d,%d]" % \
            #             (row, col, self._strandType, baseIdxLow, baseIdxHigh)
            d = "(%d,%d).%d^%d" % (row, col, self._strandType, strandSetIdx)
            util.execCommandList(self, [c], desc=d, useUndoStack=useUndoStack)
            return strandSetIdx
        else:
            return -1
    # end def

    def createDeserializedStrand(self, baseIdxLow, baseIdxHigh, useUndoStack=False):
        """
        Passes a strand to AddStrandCommand that was read in from file input.
        Omits the step of checking _couldStrandInsertAtLastIndex, since
        we assume that deserialized strands will not cause collisions.
        """
        boundsLow, boundsHigh = self.getBoundsOfEmptyRegionContaining(baseIdxLow)
        assert(baseIdxLow < baseIdxHigh)
        assert(boundsLow <= baseIdxLow)
        assert(baseIdxHigh <= boundsHigh)
        canInsert, strandSetIdx = self.getIndexToInsert(baseIdxLow, baseIdxHigh)
        if canInsert:
            c = StrandSet.CreateStrandCommand(self, baseIdxLow, baseIdxHigh, strandSetIdx)
            util.execCommandList(self, [c], desc=None, useUndoStack=useUndoStack)
            return strandSetIdx
        else:
            return -1
    # end def

    def removeStrand(self, strand, strandSetIdx=None, useUndoStack=True, solo=True):
        """
        solo is an argument to enable limiting signals emiting from
        the command in the case the command is instantiated part of a larger
        command
        """
        cmds = []
        if strandSetIdx == None:
            isInSet, overlap, strandSetIdx = self._findIndexOfRangeFor(strand)
            if not isInSet:
                raise IndexError
        if self.isScaffold() and strand.sequence() != None:
            cmds.append(strand.oligo().applySequenceCMD(None))
        cmds += strand.clearDecoratorCommands()
        cmds.append(StrandSet.RemoveStrandCommand(self, strand, strandSetIdx, solo))
        util.execCommandList(self, cmds, desc="Remove strand", useUndoStack=useUndoStack)
        return strandSetIdx
    # end def

    def removeAllStrands(self, useUndoStack=True):
        # copy the list because we are going to shrink it and that's
        # a no no with iterators
        #temp = [x for x in self._strandList]
        for strand in list(self._strandList):#temp:
            self.removeStrand(strand, 0, useUndoStack, solo=False)
        # end def

    def mergeStrands(self, priorityStrand, otherStrand, useUndoStack=True):
        """
        Merge the priorityStrand and otherStrand into a single new strand.
        The oligo of priority should be propagated to the other and all of
        its connections.
        """
        lowAndHighStrands = self.strandsCanBeMerged(priorityStrand, otherStrand)
        if lowAndHighStrands:
            strandLow, strandHigh = lowAndHighStrands
            isInSet, overlap, lowStrandSetIdx = self._findIndexOfRangeFor(strandLow)
            if isInSet:
                c = StrandSet.MergeCommand(strandLow, strandHigh, \
                                            lowStrandSetIdx, priorityStrand)
                util.execCommandList(self, [c], desc="Merge", useUndoStack=useUndoStack)
    # end def

    def strandsCanBeMerged(self, strandA, strandB):
        """
        returns None if the strands can't be merged, otherwise
        if the strands can be merge it returns the strand with the lower index

        only checks that the strands are of the same StrandSet and that the
        end points differ by 1.  DOES NOT check if the Strands overlap, that
        should be handled by addStrand
        """
        if strandA.strandSet() != strandB.strandSet():
            return None
        if abs(strandA.lowIdx() - strandB.highIdx()) == 1 or \
            abs(strandB.lowIdx() - strandA.highIdx()) == 1:
            if strandA.lowIdx() < strandB.lowIdx():
                if not strandA.connectionHigh() and not strandB.connectionLow():
                    return strandA, strandB
            else:
                if not strandB.connectionHigh() and not strandA.connectionLow():
                    return strandB, strandA
        else:
            return None
    # end def

    def splitStrand(self, strand, baseIdx, updateSequence=True, useUndoStack=True):
        """
        Break strand into two strands. Reapply sequence by default (disabled
        during autostaple).
        """
        if self.strandCanBeSplit(strand, baseIdx):
            isInSet, overlap, strandSetIdx = self._findIndexOfRangeFor(strand)
            if isInSet:
                c = StrandSet.SplitCommand(strand, baseIdx, strandSetIdx, updateSequence)
                util.execCommandList(self, [c], desc="Split", useUndoStack=useUndoStack)
                return True
            else:
                return False
        else:
            return False
    # end def

    def strandCanBeSplit(self, strand, baseIdx):
        """
        Make sure the base index is within the strand
        Don't split right next to a 3Prime end
        Don't split on endpoint (AKA a crossover)
        """
        # no endpoints
        if baseIdx == strand.lowIdx() or baseIdx == strand.highIdx():
            return False
        # make sure the base index within the strand
        elif strand.lowIdx() > baseIdx or baseIdx > strand.highIdx():
            return False
        elif abs(baseIdx - strand.idx3Prime()) > 1:
            return True
    # end def

    def destroy(self):
        self.setParent(None)
        self.deleteLater()  # QObject will emit a destroyed() Signal
    # end def

    def remove(self, useUndoStack=True):
        """
        Removes a VirtualHelix from the model. Accepts a reference to the
        VirtualHelix, or a (row,col) lattice coordinate to perform a lookup.
        """
        if useUndoStack:
            self.undoStack().beginMacro("Delete StrandSet")
        self.removeAllStrands(useUndoStack)
        if useUndoStack:
            self.undoStack().endMacro()
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def undoStack(self):
        if self._undoStack == None:
            self._undoStack = self._virtualHelix.undoStack()
        return self._undoStack

    def virtualHelix(self):
        return self._virtualHelix

    def strandFilter(self):
        return "scaffold" if self._strandType == StrandType.Scaffold else "staple"

    def hasStrandAt(self, idxLow, idxHigh):
        """
        """
        dummyStrand = Strand(self, idxLow, idxHigh)
        strandList = [s for s in self._findOverlappingRanges(dummyStrand)]
        dummyStrand._strandSet = None
        dummyStrand.setParent(None)
        dummyStrand.deleteLater()
        dummyStrand = None
        return len(strandList) > 0
    # end def

    def getOverlappingStrands(self, idxLow, idxHigh):
        dummyStrand = Strand(self, idxLow, idxHigh)
        strandList = [s for s in self._findOverlappingRanges(dummyStrand)]
        dummyStrand._strandSet = None
        dummyStrand.setParent(None)
        dummyStrand.deleteLater()
        dummyStrand = None
        return strandList
    # end def

    def hasStrandAtAndNoXover(self, idx):
        dummyStrand = Strand(self, idx, idx)
        strandList = [s for s in self._findOverlappingRanges(dummyStrand)]
        dummyStrand._strandSet = None
        dummyStrand.setParent(None)
        dummyStrand.deleteLater()
        dummyStrand = None
        if len(strandList) > 0:
            return False if strandList[0].hasXoverAt(idx) else True
        else:
            return False
    # end def

    def hasNoStrandAtOrNoXover(self, idx):
        dummyStrand = Strand(self, idx, idx)
        strandList = [s for s in self._findOverlappingRanges(dummyStrand)]
        dummyStrand._strandSet = None
        dummyStrand.setParent(None)
        dummyStrand.deleteLater()
        dummyStrand = None
        if len(strandList) > 0:
            return False if strandList[0].hasXoverAt(idx) else True
        else:
            return True
    # end def

    def getIndexToInsert(self, idxLow, idxHigh):
        """
        """
        canInsert = True
        dummyStrand = Strand(self, idxLow, idxHigh)
        if self._couldStrandInsertAtLastIndex(dummyStrand):
            return canInsert, self._lastStrandSetIndex
        isInSet, overlap, idx = self._findIndexOfRangeFor(dummyStrand)
        dummyStrand._strandSet = None
        dummyStrand.setParent(None)
        dummyStrand.deleteLater()
        dummyStrand = None
        if overlap:
            canInsert = False
        return canInsert, idx
    # end def

    def getStrand(self, baseIdx):
        """Returns the strand that overlaps with baseIdx."""
        dummyStrand = Strand(self, baseIdx, baseIdx)
        strandList = [s for s in self._findOverlappingRanges(dummyStrand)]
        dummyStrand._strandSet = None
        dummyStrand.setParent(None)
        dummyStrand.deleteLater()
        dummyStrand = None
        return strandList[0] if len(strandList) > 0 else None
    # end def

    def getLegacyArray(self):
        """docstring for getLegacyArray"""
        num = self._virtualHelix.number()
        ret = [[-1, -1, -1, -1] for i in range(self.part().maxBaseIdx() + 1)]
        if self.isDrawn5to3():
            for strand in self._strandList:
                lo, hi = strand.idxs()
                assert strand.idx5Prime() == lo and strand.idx3Prime() == hi
                # map the first base (5' xover if necessary)
                s5p = strand.connection5p()
                if s5p != None:
                    ret[lo][0] = s5p.virtualHelix().number()
                    ret[lo][1] = s5p.idx3Prime()
                ret[lo][2] = num
                ret[lo][3] = lo + 1
                # map the internal bases
                for idx in range(lo + 1, hi):
                    ret[idx][0] = num
                    ret[idx][1] = idx - 1
                    ret[idx][2] = num
                    ret[idx][3] = idx + 1
                # map the last base (3' xover if necessary)
                ret[hi][0] = num
                ret[hi][1] = hi - 1
                s3p = strand.connection3p()
                if s3p != None:
                    ret[hi][2] = s3p.virtualHelix().number()
                    ret[hi][3] = s3p.idx5Prime()
                # end if
            # end for
        # end if
        else:
            for strand in self._strandList:
                lo, hi = strand.idxs()
                assert strand.idx3Prime() == lo and strand.idx5Prime() == hi
                # map the first base (3' xover if necessary)
                ret[lo][0] = num
                ret[lo][1] = lo + 1
                s3p = strand.connection3p()
                if s3p != None:
                    ret[lo][2] = s3p.virtualHelix().number()
                    ret[lo][3] = s3p.idx5Prime()
                # map the internal bases
                for idx in range(lo + 1, hi):
                    ret[idx][0] = num
                    ret[idx][1] = idx + 1
                    ret[idx][2] = num
                    ret[idx][3] = idx - 1
                # map the last base (5' xover if necessary)
                ret[hi][2] = num
                ret[hi][3] = hi - 1
                s5p = strand.connection5p()
                if s5p != None:
                    ret[hi][0] = s5p.virtualHelix().number()
                    ret[hi][1] = s5p.idx3Prime()
                # end if
            # end for
        return ret
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _addToStrandList(self, strand, idx):
        """Inserts strand into the _strandList at idx."""
        self._strandList.insert(idx, strand)

    def _removeFromStrandList(self, strand):
        """Remove strand from _strandList."""
        self._doc.removeStrandFromSelection(strand)  # make sure the strand is no longer selected
        self._strandList.remove(strand)

    def _couldStrandInsertAtLastIndex(self, strand):
        """Verification of insertability based on cached last index."""
        lastInd = self._lastStrandSetIndex
        strandList = self._strandList
        if lastInd == None or lastInd > (len(strandList) - 1):
            self._lastStrandSetIndex = None
            return False
        else:
            sTestHigh = strandList[lastInd].lowIdx() if lastInd < len(strandList) else self.partMaxBaseIdx()
            sTestLow = strandList[lastInd - 1].highIdx() if lastInd > 0 else - 1
            sLow, sHigh = strand.idxs()
            if sTestLow < sLow and sHigh < sTestHigh:
                return True
            else:
                return False

    def _findOverlappingRanges(self, qstrand, useCache=False):
        """
        a binary search for the strands in self._strandList overlapping with
        a query strands, or qstrands, indices.

        Useful for operations on complementary strands such as applying a
        sequence

        This is an generator for now

        Strategy:
        1.
            search the _strandList for a strand the first strand that has a
            highIndex >= lowIndex of the query strand.
            save that strandSet index as sSetIndexLow.
            if No strand satisfies this condition, return an empty list

            Unless it matches the query strand's lowIndex exactly,
            Step 1 is O(log N) where N in length of self._strandList to the max,
            that is it needs to exhaust the search

            conversely you could search for first strand that has a
            lowIndex LESS than or equal to the lowIndex of the query strand.

        2.
            starting at self._strandList[sSetIndexLow] test each strand to see if
            it's indexLow is LESS than or equal to qstrand.indexHigh.  If it is
            yield/return that strand.  If it's GREATER than the indexHigh, or
            you run out of strands to check, the generator terminates
        """
        strandList = self._strandList
        lenStrands = len(strandList)
        if lenStrands == 0:
            return
        # end if

        low = 0
        high = lenStrands
        qLow, qHigh = qstrand.idxs()

        # Step 1: get rangeIndexLow with a binary search
        if useCache:  # or self.doesLastSetIndexMatch(qstrand, strandList):
            # cache match!
            sSetIndexLow = self._lastStrandSetIndex
        else:
            sSetIndexLow = -1
            while low < high:
                mid = (low + high) // 2
                midStrand = strandList[mid]

                # pre get indices from the currently tested strand
                mLow, mHigh = midStrand.idxs()

                if mHigh == qLow:
                    # match, break out of while loop
                    sSetIndexLow = mid
                    break
                elif mHigh > qLow:
                    # store the candidate index
                    sSetIndexLow = mid
                    # adjust the high index to find a better candidate if
                    # it exists
                    high = mid
                # end elif
                else:  # mHigh < qLow
                    # If a strand exists it must be a higher rangeIndex
                    # leave the high the same
                    low = mid + 1
                #end elif
            # end while
        # end else

        # Step 2: create a generator on matches
        # match on whether the tempStrand's lowIndex is
        # within the range of the qStrand
        if sSetIndexLow > -1:
            tempStrands = iter(strandList[sSetIndexLow:])
            tempStrand = next(tempStrands)
            qHigh += 1  # bump it up for a more efficient comparison
            i = 0   # use this to
            while tempStrand and tempStrand.lowIdx() < qHigh:
                yield tempStrand
                # use a next and a default to cause a break condition
                tempStrand = next(tempStrands, None)
                i += 1
            # end while

            # cache the last index we left of at
            i = sSetIndexLow + i
            """
            if
            1. we ran out of strands to test adjust
                OR
            2. the end condition tempStrands highIndex is still inside the
            qstrand but not equal to the end point
                adjust i down 1
            otherwise
            """
            if not tempStrand or tempStrand.highIdx() < qHigh - 1:
                i -= 1
            # assign cache but double check it's a valid index
            self._lastStrandSetIndex = i if -1 < i < lenStrands else None
            return
        else:
            # no strand was found
            # go ahead and clear the cache
            self._lastStrandSetIndex = None if len(self._strandList) > 0 else 0
            return
    # end def

    def getStrandIndex(self, strand):
        try:
            ind = self._strandList.index(strand)
            return (True, ind)
        except ValueError:
            return (False, 0)
    # end def

    def _findIndexOfRangeFor(self, strand):
        """
        Performs a binary search for strand in self._strandList.

        If the strand is found, we want to return its index and we don't care
        about whether it overlaps with anything.

        If the strand is not found, we want to return whether it would
        overlap with any existing strands, and if not, the index where it
        would go.

        Returns a tuple (found, overlap, idx)
            found is True if strand in self._strandList
            overlap is True if strand is not found, and would overlap with
            existing strands in self._strandList
            idx is the index where the strand was found if found is True
            idx is the index where the strand could be inserted if found
            is False and overlap is False.
        """
        # setup
        strandList = self._strandList
        lastIdx = self._lastStrandSetIndex
        lenStrands = len(strandList)
        # base case: empty list, can insert at 0
        if lenStrands == 0:
            return (False, False, 0)
        # check cache
        if lastIdx:
            if lastIdx < lenStrands and strandList[lastIdx] == strand:
                return (True, False, lastIdx)
        # init search bounds
        low, high = 0, lenStrands
        sLow, sHigh = strand.idxs()
        # perform binary search
        while low < high:
            mid = (low + high) // 2
            midStrand = strandList[mid]
            mLow, mHigh = midStrand.idxs()
            if midStrand == strand:
                self._lastStrandSetIndex = mid
                return (True, False, mid)
            elif mHigh < sLow:
                #  strand                [sLow----)
                # mStrand  (----mHigh]
                low = mid + 1  # search higher
            elif mLow > sHigh:
                #  strand  (----sHigh]
                # mStrand                [mLow----)
                high = mid  # search lower
            else:
                if mLow <= sLow <= mHigh:
                    # overlap: right side of mStrand
                    #  strand         [sLow---------------)
                    # mStrand  [mLow----------mHigh]
                    self._lastStrandSetIndex = None
                    return (False, True, None)
                elif mLow <= sHigh <= mHigh:
                    # overlap: left side of mStrand
                    #  strand  (--------------sHigh]
                    # mStrand         [mLow-----------mHigh]
                    self._lastStrandSetIndex = None
                    return (False, True, None)
                elif sLow <= mLow and mHigh <= sHigh:
                    # overlap: strand encompases existing
                    #  strand  [sLow-------------------sHigh]
                    # mStrand         [mLow----mHigh]
                    # note: inverse case is already covered above
                    self._lastStrandSetIndex = None
                    return (False, True, None)
                else:
                    # strand not in set, here's where you'd insert it
                    self._lastStrandSetIndex = mid
                    return (False, False, mid)
            # end else
        self._lastStrandSetIndex = low
        return (False, False, low)
    # end def

    def _doesLastSetIndexMatch(self, qstrand, strandList):
        """
        strandList is passed to save a lookup
        """
        lSI = self._lastStrandSetIndex
        if lSI:
            qLow, qHigh = qstrand.idxs()
            tempStrand = strandList[lSI]
            tLow, tHigh = tempStrand.idxs()
            if not (qLow <= tLow <= qHigh or qLow <= tHigh <= qHigh):
                return False
            else:  # get a difference
                dif = abs(qLow - tLow)
                # check neighboring strandList just in case
                difLow = dif + 1
                if lSI > 0:
                    tLow, tHigh = strand[lSI - 1].idxs()
                    if qLow <= tLow <= qHigh or qLow <= tHigh <= qHigh:
                        difLow = abs(qLow - tLow)
                difHigh = dif + 1
                if lSI < len(strand) - 1:
                    tLow, tHigh = strand[lSI + 1].idxs()
                    if qLow <= tLow <= qHigh or qLow <= tHigh <= qHigh:
                        difHigh = abs(qLow - tLow)
                # check that the cached strand is in fact the right guy
                if dif < difLow and dif < difHigh:
                    return True
                else:
                    False
            # end else
        # end if
        else:
            return False
        # end else
    # end def

    ### COMMANDS ###
    class CreateStrandCommand(QUndoCommand):
        """
        Create a new Strand based with bounds (baseIdxLow, baseIdxHigh),
        and insert it into the strandSet at position strandSetIdx. Also,
        create a new Oligo, add it to the Part, and point the new Strand
        at the oligo.
        """
        def __init__(self, strandSet, baseIdxLow, baseIdxHigh, strandSetIdx):
            super(StrandSet.CreateStrandCommand, self).__init__()
            self._strandSet = strandSet
            self._sSetIdx = strandSetIdx
            self._strand = Strand(strandSet, baseIdxLow, baseIdxHigh)
            colorList = styles.stapColors if strandSet.isStaple() else styles.scafColors
            color = random.choice(colorList).name()
            self._newOligo = Oligo(None, color)  # redo will set part
            self._newOligo.setLength(self._strand.totalLength())
        # end def

        def redo(self):
            # Add the new strand to the StrandSet strandList
            strand = self._strand
            strandSet = self._strandSet
            strandSet._strandList.insert(self._sSetIdx, strand)
            # Set up the new oligo
            oligo = self._newOligo
            oligo.setStrand5p(strand)
            oligo.addToPart(strandSet.part())
            strand.setOligo(oligo)

            if strandSet.isStaple():
                strand.reapplySequence()
            # Emit a signal to notify on completion
            strandSet.strandsetStrandAddedSignal.emit(strandSet, strand)
            # for updating the Slice View displayed helices
            strandSet.part().partStrandChangedSignal.emit(strandSet.part(), strandSet.virtualHelix())
        # end def

        def undo(self):
            # Remove the strand from StrandSet strandList and selectionList
            strand = self._strand
            strandSet = self._strandSet
            strandSet._doc.removeStrandFromSelection(strand)
            strandSet._strandList.pop(self._sSetIdx)
            # Get rid of the new oligo
            oligo = self._newOligo
            oligo.setStrand5p(None)
            oligo.removeFromPart()
            # Emit a signal to notify on completion
            strand.strandRemovedSignal.emit(strand)
            strand.setOligo(None)
            # for updating the Slice View displayed helices
            strandSet.part().partStrandChangedSignal.emit(strandSet.part(), strandSet.virtualHelix())
        # end def
    # end class

    class RemoveStrandCommand(QUndoCommand):
        """
        RemoveStrandCommand deletes a strand. It should only be called on
        strands with no connections to other strands.
        """
        def __init__(self, strandSet, strand, strandSetIdx, solo=True):
            super(StrandSet.RemoveStrandCommand, self).__init__()
            self._strandSet = strandSet
            self._strand = strand
            self._sSetIdx = strandSetIdx
            self._solo = solo
            self._oldStrand5p = strand.connection5p()
            self._oldStrand3p = strand.connection3p()
            self._oligo = olg = strand.oligo()
            # only create a new 5p oligo if there is a 3' connection
            self._newOligo5p = olg.shallowCopy() if self._oldStrand5p else None
            if olg.isLoop() or self._oldStrand3p == None:
                self._newOligo3p = olg3p = None
                if self._newOligo5p:
                    self._newOligo5p.setLoop(False)
            else:
                self._newOligo3p = olg3p = olg.shallowCopy()
                olg3p.setStrand5p(self._oldStrand3p)
                colorList = styles.stapColors if strandSet.isStaple() else styles.scafColors
                color = random.choice(colorList).name()
                olg3p.setColor(color)
                olg3p.refreshLength()
        # end def

        def redo(self):
            # Remove the strand
            strand = self._strand
            strandSet = self._strandSet
            # strandSet._removeFromStrandList(strand)
            strandSet._doc.removeStrandFromSelection(strand)
            strandSet._strandList.pop(self._sSetIdx)
            strand5p = self._oldStrand5p
            strand3p = self._oldStrand3p
            oligo = self._oligo
            olg5p = self._newOligo5p
            olg3p = self._newOligo3p

            #oligo.incrementLength(-strand.totalLength())
            
            oligo.removeFromPart()

            if strand5p != None:
                strand5p.setConnection3p(None)
            if strand3p != None:
                strand3p.setConnection5p(None)

            # Clear connections and update oligos
            if strand5p != None:
                for s5p in oligo.strand5p().generator3pStrand():
                    Strand.setOligo(s5p, olg5p)
                olg5p.refreshLength()
                olg5p.addToPart(strandSet.part())
                if self._solo:
                    part = strandSet.part()
                    vh = strandSet.virtualHelix()
                    part.partActiveVirtualHelixChangedSignal.emit(part, vh)
                    #strand5p.strandXover5pChangedSignal.emit(strand5p, strand)
                strand5p.strandUpdateSignal.emit(strand5p)
            # end if
            if strand3p != None:
                if not oligo.isLoop():
                    # apply 2nd oligo copy to all 3' downstream strands
                    for s3p in strand3p.generator3pStrand():
                        Strand.setOligo(s3p, olg3p)
                    olg3p.addToPart(strandSet.part())
                if self._solo:
                    part = strandSet.part()
                    vh = strandSet.virtualHelix()
                    part.partActiveVirtualHelixChangedSignal.emit(part, vh)
                    # strand.strandXover5pChangedSignal.emit(strand, strand3p)
                strand3p.strandUpdateSignal.emit(strand3p)
            # end if
            # Emit a signal to notify on completion
            strand.strandRemovedSignal.emit(strand)
            # for updating the Slice View displayed helices
            strandSet.part().partStrandChangedSignal.emit(strandSet.part(), strandSet.virtualHelix())
        # end def

        def undo(self):
            # Restore the strand
            strand = self._strand
            strandSet = self._strandSet
            # Add the newStrand to the sSet
            strandSet._addToStrandList(strand, self._sSetIdx)
            # strandSet._strandList.insert(self._sSetIdx, strand)
            strand5p = self._oldStrand5p
            strand3p = self._oldStrand3p
            oligo = self._oligo
            olg5p = self._newOligo5p
            olg3p = self._newOligo3p

            # Restore connections to this strand
            if strand5p != None:
                strand5p.setConnection3p(strand)

            if strand3p != None:
                strand3p.setConnection5p(strand)

            # oligo.decrementLength(strand.totalLength())
            
            # Restore the oligo
            oligo.addToPart(strandSet.part())
            if olg5p:
                olg5p.removeFromPart()
            if olg3p:
                olg3p.removeFromPart()
            for s5p in oligo.strand5p().generator3pStrand():
                Strand.setOligo(s5p, oligo)
            # end for

            # Emit a signal to notify on completion
            strandSet.strandsetStrandAddedSignal.emit(strandSet, strand)
            # for updating the Slice View displayed helices
            strandSet.part().partStrandChangedSignal.emit(strandSet.part(), strandSet.virtualHelix())

            # Restore connections to this strand
            if strand5p != None:
                if self._solo:
                    part = strandSet.part()
                    vh = strandSet.virtualHelix()
                    part.partActiveVirtualHelixChangedSignal.emit(part, vh)
                    # strand5p.strandXover5pChangedSignal.emit(
                    #                                        strand5p, strand)
                strand5p.strandUpdateSignal.emit(strand5p)
                strand.strandUpdateSignal.emit(strand)

            if strand3p != None:
                if self._solo:
                    part = strandSet.part()
                    vh = strandSet.virtualHelix()
                    part.partActiveVirtualHelixChangedSignal.emit(part, vh)
                    # strand.strandXover5pChangedSignal.emit(strand, strand3p)
                strand3p.strandUpdateSignal.emit(strand3p)
                strand.strandUpdateSignal.emit(strand)
        # end def
    # end class

    class MergeCommand(QUndoCommand):
        """
        This class takes two Strands and merges them.  This Class should be
        private to StrandSet as knowledge of a strandSetIndex outside of this
        of the StrandSet class implies knowledge of the StrandSet
        implementation

        Must pass this two different strands, and nominally one of the strands
        again which is the priorityStrand.  The resulting "merged" strand has
        the properties of the priorityStrand's oligo.  Decorators are preserved

        the strandLow and strandHigh must be presorted such that strandLow
        has a lower range than strandHigh

        lowStrandSetIdx should be known ahead of time as a result of selection
        """
        def __init__(self, strandLow, strandHigh, lowStrandSetIdx, priorityStrand):
            super(StrandSet.MergeCommand, self).__init__()
            # Store strands
            self._strandLow = strandLow
            self._strandHigh = strandHigh
            pS = priorityStrand
            self._sSet = sSet = pS.strandSet()
            # Store oligos
            self._newOligo = pS.oligo().shallowCopy()
            self._sLowOligo = sLOlg = strandLow.oligo()
            self._sHighOligo = sHOlg = strandHigh.oligo()

            self._sSetIdx = lowStrandSetIdx

            # update the new oligo length if it's not a loop
            if sLOlg != sHOlg:
                self._newOligo.setLength(sLOlg.length() + sHOlg.length())

            # Create the newStrand by copying the priority strand to
            # preserve its properties
            newIdxs = strandLow.lowIdx(), strandHigh.highIdx()
            newStrand = strandLow.shallowCopy()
            newStrand.setIdxs(newIdxs)
            newStrand.setConnectionHigh(strandHigh.connectionHigh())
            # Merging any decorators
            newStrand.addDecorators(strandHigh.decorators())
            self._newStrand = newStrand
            # Update the oligo for things like its 5prime end and isLoop
            self._newOligo.strandMergeUpdate(strandLow, strandHigh, newStrand)
            
            # set the new sequence by concatenating the sequence properly
            if strandLow._sequence or strandHigh._sequence:
                tL = strandLow.totalLength()
                tH = strandHigh.totalLength()
                seqL = strandLow._sequence if strandLow._sequence else "".join([" " for i in range(tL)])
                seqH = strandHigh._sequence if strandHigh._sequence else "".join([" " for i in range(tH)])    
                if newStrand.isDrawn5to3():
                    newStrand._sequence = seqL + seqH
                else:
                    newStrand._sequence = seqH + seqL
        # end def

        def redo(self):
            sS = self._sSet
            sL = self._strandLow
            sH = self._strandHigh
            nS = self._newStrand
            idx = self._sSetIdx
            olg = self._newOligo
            lOlg = sL.oligo()
            hOlg = sH.oligo()

            # Remove old strands from the sSet (reusing idx, so order matters)
            sS._removeFromStrandList(sL)
            sS._removeFromStrandList(sH)
            # Add the newStrand to the sSet
            sS._addToStrandList(nS, idx)

            # update connectivity of strands
            nScL = nS.connectionLow()
            if nScL:
                if (nS.isDrawn5to3() and nScL.isDrawn5to3()) or \
                    (not nS.isDrawn5to3() and not nScL.isDrawn5to3()):
                    nScL.setConnectionHigh(nS)
                else:
                    nScL.setConnectionLow(nS)
            nScH = nS.connectionHigh()
            if nScH:
                if (nS.isDrawn5to3() and nScH.isDrawn5to3()) or \
                    (not nS.isDrawn5to3() and not nScH.isDrawn5to3()):
                    nScH.setConnectionLow(nS)
                else:
                    nScH.setConnectionHigh(nS)

            # Traverse the strands via 3'conns to assign the new oligo
            for strand in olg.strand5p().generator3pStrand():
                Strand.setOligo(strand, olg)  # emits strandHasNewOligoSignal

            # Add new oligo and remove old oligos
            olg.addToPart(sS.part())
            lOlg.removeFromPart()
            if hOlg != lOlg:  # check if a loop was created
                hOlg.removeFromPart()

            # Emit Signals related to destruction and addition
            sL.strandRemovedSignal.emit(sL)
            sH.strandRemovedSignal.emit(sH)
            sS.strandsetStrandAddedSignal.emit(sS, nS)
        # end def

        def undo(self):
            sS = self._sSet
            sL = self._strandLow
            sH = self._strandHigh
            nS = self._newStrand
            idx = self._sSetIdx
            olg = self._newOligo
            lOlg = self._sLowOligo
            hOlg = self._sHighOligo
            # Remove the newStrand from the sSet
            sS._removeFromStrandList(nS)
            # Add old strands to the sSet (reusing idx, so order matters)
            sS._addToStrandList(sH, idx)
            sS._addToStrandList(sL, idx)

            # update connectivity of strands
            sLcL = sL.connectionLow()
            if sLcL:
                if (sL.isDrawn5to3() and sLcL.isDrawn5to3()) or \
                    (not sL.isDrawn5to3() and not sLcL.isDrawn5to3()):
                    sLcL.setConnectionHigh(sL)
                else:
                    sLcL.setConnectionLow(sL)
            sHcH = sH.connectionHigh()
            if sHcH:
                if (sH.isDrawn5to3() and sHcH.isDrawn5to3()) or \
                    (not sH.isDrawn5to3() and not sHcH.isDrawn5to3()):
                    sHcH.setConnectionLow(sH)
                else:
                    sHcH.setConnectionHigh(sH)

            # Traverse the strands via 3'conns to assign the old oligo
            for strand in lOlg.strand5p().generator3pStrand():
                Strand.setOligo(strand, lOlg)  # emits strandHasNewOligoSignal
            for strand in hOlg.strand5p().generator3pStrand():
                Strand.setOligo(strand, hOlg)  # emits strandHasNewOligoSignal

            # Remove new oligo and add old oligos
            olg.removeFromPart()
            lOlg.addToPart(sL.part())
            if hOlg != lOlg:
                hOlg.addToPart(sH.part())

            # Emit Signals related to destruction and addition
            nS.strandRemovedSignal.emit(nS)
            sS.strandsetStrandAddedSignal.emit(sS, sL)
            sS.strandsetStrandAddedSignal.emit(sS, sH)
        # end def
    # end class

    class SplitCommand(QUndoCommand):
        """
        The SplitCommand takes as input a strand and "splits" the strand in
        two, such that one new strand 3' end is at baseIdx, and the other
        new strand 5' end is at baseIdx +/- 1 (depending on the direction
        of the strands).

        Under the hood:
        On redo, this command actually is creates two new copies of the
        original strand, resizes each and modifies their connections.
        On undo, the new copies are removed and the original is restored.
        """
        def __init__(self, strand, baseIdx, strandSetIdx, updateSequence=True):
            super(StrandSet.SplitCommand, self).__init__()
            # Store inputs
            self._oldStrand = strand
            oldSequence  = strand._sequence
            is5to3 = strand.isDrawn5to3()
            
            self._sSetIdx = strandSetIdx
            self._sSet = sSet = strand.strandSet()
            self._oldOligo = oligo = strand.oligo()
            # Create copies
            self._strandLow = strandLow = strand.shallowCopy()
            self._strandHigh = strandHigh = strand.shallowCopy()

            if oligo.isLoop():
                self._lOligo = self._hOligo = lOligo = hOligo = oligo.shallowCopy()
            else:
                self._lOligo = lOligo = oligo.shallowCopy()
                self._hOligo = hOligo = oligo.shallowCopy()
            colorList = styles.stapColors if sSet.isStaple() \
                                            else styles.scafColors
            # end

            # Determine oligo retention based on strand priority
            if is5to3:  # strandLow has priority
                iNewLow = baseIdx
                colorLow = oligo.color()
                colorHigh = random.choice(colorList).name()
                olg5p, olg3p = lOligo, hOligo
                std5p, std3p = strandLow, strandHigh
            else:  # strandHigh has priority
                iNewLow = baseIdx - 1
                colorLow = random.choice(colorList).name()
                colorHigh = oligo.color()
                olg5p, olg3p = hOligo, lOligo
                std5p, std3p = strandHigh, strandLow
            # this is for updating a connected xover view object
            # there is only ever one xover a strand is in charge of
            self._strand3p = std3p
            self._strand5p = std5p

            # Update strand connectivity
            strandLow.setConnectionHigh(None)
            strandHigh.setConnectionLow(None)

            # Resize strands and update decorators
            strandLow.setIdxs((strand.lowIdx(), iNewLow))
            strandHigh.setIdxs((iNewLow + 1, strand.highIdx()))

            # Update the oligo for things like its 5prime end and isLoop
            olg5p.strandSplitUpdate(std5p, std3p, olg3p, strand)

            if not oligo.isLoop():
                # Update the oligo color if necessary
                lOligo.setColor(colorLow)
                hOligo.setColor(colorHigh)
                # settle the oligo length
                length = 0
                for strand in std3p.generator3pStrand():
                    length += strand.totalLength()
                # end for
                olg5p.setLength(olg5p.length() - length)
                olg3p.setLength(length)
            # end if

            if updateSequence and oldSequence:
                if is5to3:  # strandLow has priority
                    tL = strandLow.totalLength()
                    strandLow._sequence = oldSequence[0:tL]
                    strandHigh._sequence = oldSequence[tL:]
                    # print "lenght match 5 to 3", strandHigh.totalLength()+tL == len(oldSequence)
                    # assert (strandHigh.totalLength()+tL == len(oldSequence))
                else:
                    tH = strandHigh.totalLength()
                    strandHigh._sequence = oldSequence[0:tH]
                    strandLow._sequence = oldSequence[tH:]
                    # print "lenght match 3 to 5", strandLow.totalLength()+tH == len(oldSequence)
                    # assert (strandLow.totalLength()+tH == len(oldSequence))

        # end def

        def redo(self):
            sS = self._sSet
            sL = self._strandLow
            sH = self._strandHigh
            oS = self._oldStrand
            idx = self._sSetIdx
            olg = self._oldOligo
            lOlg = self._lOligo
            hOlg = self._hOligo
            wasNotLoop = lOlg != hOlg

            # Remove old Strand from the sSet
            sS._removeFromStrandList(oS)

            # Add new strands to the sSet (reusing idx, so order matters)
            sS._addToStrandList(sH, idx)
            sS._addToStrandList(sL, idx)

            # update connectivity of strands
            sLcL = sL.connectionLow()
            if sLcL:
                if (oS.isDrawn5to3() and sLcL.isDrawn5to3()) or \
                    (not oS.isDrawn5to3() and not sLcL.isDrawn5to3()):
                    sLcL.setConnectionHigh(sL)
                else:
                    sLcL.setConnectionLow(sL)
            sHcH = sH.connectionHigh()
            if sHcH:
                if (oS.isDrawn5to3() and sHcH.isDrawn5to3()) or \
                    (not oS.isDrawn5to3() and not sHcH.isDrawn5to3()):
                    sHcH.setConnectionLow(sH)
                else:
                    sHcH.setConnectionHigh(sH)

            # Traverse the strands via 3'conns to assign the new oligos
            for strand in lOlg.strand5p().generator3pStrand():
                Strand.setOligo(strand, lOlg)  # emits strandHasNewOligoSignal
            if wasNotLoop:  # do the second oligo which is different
                for strand in hOlg.strand5p().generator3pStrand():
                    # emits strandHasNewOligoSignal
                    Strand.setOligo(strand, hOlg)

            # Add new oligo and remove old oligos from the part
            olg.removeFromPart()
            lOlg.addToPart(sL.part())
            if wasNotLoop:
                hOlg.addToPart(sH.part())

            # Emit Signals related to destruction and addition
            oS.strandRemovedSignal.emit(oS)
            sS.strandsetStrandAddedSignal.emit(sS, sH)
            sS.strandsetStrandAddedSignal.emit(sS, sL)
        # end def

        def undo(self):
            sS = self._sSet
            sL = self._strandLow
            sH = self._strandHigh
            oS = self._oldStrand
            idx = self._sSetIdx
            olg = self._oldOligo
            lOlg = self._lOligo
            hOlg = self._hOligo
            wasNotLoop = lOlg != hOlg

            # Remove new strands from the sSet (reusing idx, so order matters)
            sS._removeFromStrandList(sL)
            sS._removeFromStrandList(sH)
            # Add the old strand to the sSet
            sS._addToStrandList(oS, idx)

            # update connectivity of strands
            oScL = oS.connectionLow()
            if oScL:
                if (oS.isDrawn5to3() and oScL.isDrawn5to3()) or \
                    (not oS.isDrawn5to3() and not oScL.isDrawn5to3()):
                    oScL.setConnectionHigh(oS)
                else:
                    oScL.setConnectionLow(oS)
            oScH = oS.connectionHigh()
            if oScH:
                if (oS.isDrawn5to3() and oScH.isDrawn5to3()) or \
                    (not oS.isDrawn5to3() and not oScH.isDrawn5to3()):
                    oScH.setConnectionLow(oS)
                else:
                    oScH.setConnectionHigh(oS)

            # Traverse the strands via 3'conns to assign the old oligo
            for strand in olg.strand5p().generator3pStrand():
                Strand.setOligo(strand, olg)
            # Add old oligo and remove new oligos from the part
            olg.addToPart(sS.part())
            lOlg.removeFromPart()
            if wasNotLoop:
                hOlg.removeFromPart()

            # Emit Signals related to destruction and addition
            sL.strandRemovedSignal.emit(sL)
            sH.strandRemovedSignal.emit(sH)
            sS.strandsetStrandAddedSignal.emit(sS, oS)
        # end def
    # end class

    def deepCopy(self, virtualHelix):
        """docstring for deepCopy"""
        pass
    # end def
# end class



# Form implementation generated from reading ui file 'dialogs/preferences.ui'
#
# Created by: PyQt6 UI code generator 6.1.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Preferences(object):
    def setupUi(self, Preferences):
        Preferences.setObjectName("Preferences")
        Preferences.resize(465, 374)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Preferences.sizePolicy().hasHeightForWidth())
        Preferences.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(Preferences)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(Preferences)
        self.tabWidget.setObjectName("tabWidget")
        self.settings = QtWidgets.QWidget()
        self.settings.setObjectName("settings")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.settings)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.tabWidget_2 = QtWidgets.QTabWidget(self.settings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget_2.sizePolicy().hasHeightForWidth())
        self.tabWidget_2.setSizePolicy(sizePolicy)
        self.tabWidget_2.setObjectName("tabWidget_2")
        self.tab_honeycomb = QtWidgets.QWidget()
        self.tab_honeycomb.setObjectName("tab_honeycomb")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tab_honeycomb)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.formLayout_3 = QtWidgets.QFormLayout()
        self.formLayout_3.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.FieldsStayAtSizeHint)
        self.formLayout_3.setContentsMargins(-1, 0, -1, 0)
        self.formLayout_3.setObjectName("formLayout_3")
        self.dimensionsLabel1 = QtWidgets.QLabel(self.tab_honeycomb)
        self.dimensionsLabel1.setObjectName("dimensionsLabel1")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.dimensionsLabel1)
        self.honeycombGLayout = QtWidgets.QGridLayout()
        self.honeycombGLayout.setContentsMargins(-1, 3, -1, -1)
        self.honeycombGLayout.setObjectName("honeycombGLayout")
        self.honeycombRowsSpinBox = QtWidgets.QSpinBox(self.tab_honeycomb)
        self.honeycombRowsSpinBox.setMinimum(6)
        self.honeycombRowsSpinBox.setMaximum(200)
        self.honeycombRowsSpinBox.setProperty("value", 30)
        self.honeycombRowsSpinBox.setObjectName("honeycombRowsSpinBox")
        self.honeycombGLayout.addWidget(self.honeycombRowsSpinBox, 0, 0, 1, 1)
        self.honeycombColsSpinBox = QtWidgets.QSpinBox(self.tab_honeycomb)
        self.honeycombColsSpinBox.setMinimum(6)
        self.honeycombColsSpinBox.setMaximum(200)
        self.honeycombColsSpinBox.setProperty("value", 32)
        self.honeycombColsSpinBox.setObjectName("honeycombColsSpinBox")
        self.honeycombGLayout.addWidget(self.honeycombColsSpinBox, 0, 1, 1, 1)
        self.hRowsLabel = QtWidgets.QLabel(self.tab_honeycomb)
        self.hRowsLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.hRowsLabel.setObjectName("hRowsLabel")
        self.honeycombGLayout.addWidget(self.hRowsLabel, 1, 0, 1, 1)
        self.hColsLabel = QtWidgets.QLabel(self.tab_honeycomb)
        self.hColsLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.hColsLabel.setObjectName("hColsLabel")
        self.honeycombGLayout.addWidget(self.hColsLabel, 1, 1, 1, 1)
        self.honeycombStepsSpinBox = QtWidgets.QSpinBox(self.tab_honeycomb)
        self.honeycombStepsSpinBox.setMinimum(2)
        self.honeycombStepsSpinBox.setMaximum(2000)
        self.honeycombStepsSpinBox.setSingleStep(1)
        self.honeycombStepsSpinBox.setProperty("value", 2)
        self.honeycombStepsSpinBox.setObjectName("honeycombStepsSpinBox")
        self.honeycombGLayout.addWidget(self.honeycombStepsSpinBox, 0, 2, 1, 1)
        self.hBasesLabel = QtWidgets.QLabel(self.tab_honeycomb)
        self.hBasesLabel.setObjectName("hBasesLabel")
        self.honeycombGLayout.addWidget(self.hBasesLabel, 1, 2, 1, 1)
        self.formLayout_3.setLayout(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self.honeycombGLayout)
        self.verticalLayout_3.addLayout(self.formLayout_3)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/part/dlg_honeycomb"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.tabWidget_2.addTab(self.tab_honeycomb, icon, "")
        self.tab_square = QtWidgets.QWidget()
        self.tab_square.setObjectName("tab_square")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab_square)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setContentsMargins(-1, 0, -1, 0)
        self.formLayout_2.setObjectName("formLayout_2")
        self.dimensionsLabel2 = QtWidgets.QLabel(self.tab_square)
        self.dimensionsLabel2.setObjectName("dimensionsLabel2")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.dimensionsLabel2)
        self.squareGLayout = QtWidgets.QGridLayout()
        self.squareGLayout.setContentsMargins(-1, 3, -1, -1)
        self.squareGLayout.setObjectName("squareGLayout")
        self.squareRowsSpinBox = QtWidgets.QSpinBox(self.tab_square)
        self.squareRowsSpinBox.setMinimum(6)
        self.squareRowsSpinBox.setMaximum(200)
        self.squareRowsSpinBox.setProperty("value", 30)
        self.squareRowsSpinBox.setObjectName("squareRowsSpinBox")
        self.squareGLayout.addWidget(self.squareRowsSpinBox, 0, 0, 1, 1)
        self.squareColsSpinBox = QtWidgets.QSpinBox(self.tab_square)
        self.squareColsSpinBox.setMinimum(6)
        self.squareColsSpinBox.setMaximum(200)
        self.squareColsSpinBox.setProperty("value", 30)
        self.squareColsSpinBox.setObjectName("squareColsSpinBox")
        self.squareGLayout.addWidget(self.squareColsSpinBox, 0, 1, 1, 1)
        self.sRowsLabel = QtWidgets.QLabel(self.tab_square)
        self.sRowsLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.sRowsLabel.setObjectName("sRowsLabel")
        self.squareGLayout.addWidget(self.sRowsLabel, 1, 0, 1, 1)
        self.sColsLabel = QtWidgets.QLabel(self.tab_square)
        self.sColsLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.sColsLabel.setObjectName("sColsLabel")
        self.squareGLayout.addWidget(self.sColsLabel, 1, 1, 1, 1)
        self.squareStepsSpinBox = QtWidgets.QSpinBox(self.tab_square)
        self.squareStepsSpinBox.setMinimum(2)
        self.squareStepsSpinBox.setMaximum(2000)
        self.squareStepsSpinBox.setObjectName("squareStepsSpinBox")
        self.squareGLayout.addWidget(self.squareStepsSpinBox, 0, 2, 1, 1)
        self.sBasesLabel = QtWidgets.QLabel(self.tab_square)
        self.sBasesLabel.setObjectName("sBasesLabel")
        self.squareGLayout.addWidget(self.sBasesLabel, 1, 2, 1, 1)
        self.formLayout_2.setLayout(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self.squareGLayout)
        self.verticalLayout_2.addLayout(self.formLayout_2)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/part/dlg_square"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.tabWidget_2.addTab(self.tab_square, icon1, "")
        self.verticalLayout_4.addWidget(self.tabWidget_2)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.FieldsStayAtSizeHint)
        self.formLayout.setObjectName("formLayout")
        self.autoscafLabel = QtWidgets.QLabel(self.settings)
        self.autoscafLabel.setObjectName("autoscafLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.LabelRole, self.autoscafLabel)
        self.autoScafComboBox = QtWidgets.QComboBox(self.settings)
        self.autoScafComboBox.setObjectName("autoScafComboBox")
        self.autoScafComboBox.addItem("")
        self.autoScafComboBox.addItem("")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.FieldRole, self.autoScafComboBox)
        self.zoomSpeedLabel = QtWidgets.QLabel(self.settings)
        self.zoomSpeedLabel.setObjectName("zoomSpeedLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.ItemRole.LabelRole, self.zoomSpeedLabel)
        self.zoomSpeedSlider = QtWidgets.QSlider(self.settings)
        self.zoomSpeedSlider.setMinimumSize(QtCore.QSize(140, 0))
        self.zoomSpeedSlider.setMinimum(1)
        self.zoomSpeedSlider.setMaximum(100)
        self.zoomSpeedSlider.setSingleStep(1)
        self.zoomSpeedSlider.setProperty("value", 50)
        self.zoomSpeedSlider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.zoomSpeedSlider.setInvertedControls(False)
        self.zoomSpeedSlider.setTickPosition(QtWidgets.QSlider.TickPosition.TicksBelow)
        self.zoomSpeedSlider.setTickInterval(0)
        self.zoomSpeedSlider.setObjectName("zoomSpeedSlider")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.ItemRole.FieldRole, self.zoomSpeedSlider)
        self.defaultToolComboBox = QtWidgets.QComboBox(self.settings)
        self.defaultToolComboBox.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.defaultToolComboBox.setObjectName("defaultToolComboBox")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/pathtools/dlg_select"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.defaultToolComboBox.addItem(icon2, "")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/pathtools/dlg_pencil"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.defaultToolComboBox.addItem(icon3, "")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/pathtools/dlg_paint"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.defaultToolComboBox.addItem(icon4, "")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/pathtools/dlg_addseq"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.defaultToolComboBox.addItem(icon5, "")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.FieldRole, self.defaultToolComboBox)
        self.defaultToolLabel = QtWidgets.QLabel(self.settings)
        self.defaultToolLabel.setObjectName("defaultToolLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.defaultToolLabel)
        self.verticalLayout_4.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.settings)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.RestoreDefaults)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_4.addWidget(self.buttonBox)
        self.tabWidget.addTab(self.settings, "")
        self.plugins = QtWidgets.QWidget()
        self.plugins.setObjectName("plugins")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.plugins)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.pluginTableWidget = QtWidgets.QTableWidget(self.plugins)
        self.pluginTableWidget.setObjectName("pluginTableWidget")
        self.pluginTableWidget.setColumnCount(1)
        self.pluginTableWidget.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.pluginTableWidget.setHorizontalHeaderItem(0, item)
        self.pluginTableWidget.horizontalHeader().setStretchLastSection(True)
        self.verticalLayout_5.addWidget(self.pluginTableWidget)
        self.addPluginButton = QtWidgets.QPushButton(self.plugins)
        self.addPluginButton.setObjectName("addPluginButton")
        self.verticalLayout_5.addWidget(self.addPluginButton)
        self.tabWidget.addTab(self.plugins, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.actionClose = QtGui.QAction(Preferences)
        self.actionClose.setObjectName("actionClose")

        self.retranslateUi(Preferences)
        self.tabWidget.setCurrentIndex(0)
        self.tabWidget_2.setCurrentIndex(0)
        self.defaultToolComboBox.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Preferences)

    def retranslateUi(self, Preferences):
        _translate = QtCore.QCoreApplication.translate
        Preferences.setWindowTitle(_translate("Preferences", "Preferences"))
        self.dimensionsLabel1.setText(_translate("Preferences", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Lucida Grande\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Default</p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">dimensions</p></body></html>"))
        self.hRowsLabel.setText(_translate("Preferences", "rows"))
        self.hColsLabel.setText(_translate("Preferences", "columns"))
        self.hBasesLabel.setText(_translate("Preferences", "bases (x21)"))
        self.tabWidget_2.setTabText(self.tabWidget_2.indexOf(self.tab_honeycomb), _translate("Preferences", "Honeycomb"))
        self.dimensionsLabel2.setText(_translate("Preferences", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Lucida Grande\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Default</p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">dimensions</p></body></html>"))
        self.sRowsLabel.setText(_translate("Preferences", "rows"))
        self.sColsLabel.setText(_translate("Preferences", "columns"))
        self.sBasesLabel.setText(_translate("Preferences", "bases (x32)"))
        self.tabWidget_2.setTabText(self.tabWidget_2.indexOf(self.tab_square), _translate("Preferences", "Square"))
        self.autoscafLabel.setText(_translate("Preferences", "Auto-scaffold style:"))
        self.autoScafComboBox.setItemText(0, _translate("Preferences", "Mid-seam"))
        self.autoScafComboBox.setItemText(1, _translate("Preferences", "Raster"))
        self.zoomSpeedLabel.setText(_translate("Preferences", "Mousewheel zoom speed:"))
        self.defaultToolComboBox.setItemText(0, _translate("Preferences", "Select Tool"))
        self.defaultToolComboBox.setItemText(1, _translate("Preferences", "Pencil Tool"))
        self.defaultToolComboBox.setItemText(2, _translate("Preferences", "Paint"))
        self.defaultToolComboBox.setItemText(3, _translate("Preferences", "Add Seq"))
        self.defaultToolLabel.setText(_translate("Preferences", "Default tool at startup:"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.settings), _translate("Preferences", "Settings"))
        item = self.pluginTableWidget.horizontalHeaderItem(0)
        item.setText(_translate("Preferences", "Path"))
        self.addPluginButton.setText(_translate("Preferences", "Add Plugin"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.plugins), _translate("Preferences", "Plugins"))
        self.actionClose.setText(_translate("Preferences", "Close"))
        self.actionClose.setShortcut(_translate("Preferences", "Ctrl+W"))



"""
activeslicehandle.py
Created by Shawn on 2011-02-05.
"""

from cadnano2.controllers.itemcontrollers.activesliceitemcontroller import ActiveSliceItemController
import cadnano2.util as util

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt', 'QObject',\
                                        'pyqtSignal', 'pyqtSlot', 'QEvent'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QFont',
                                       'QPen',
                                       'QDrag',
                                       'QUndoCommand'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsRectItem',
                                           'QGraphicsSimpleTextItem'])


class ActiveSliceItem(QGraphicsRectItem):
    """ActiveSliceItem for the Slice View"""
    def __init__(self, partItem, activeBaseIndex):
        super(ActiveSliceItem, self).__init__(partItem)
        self._partItem = partItem
        self._controller = ActiveSliceItemController(self, partItem.part())
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents)
    # end def

    ### SLOTS ###
    def strandChangedSlot(self, sender, vh):
        if vh == None:
            return
        partItem = self._partItem
        vhi = partItem.getVirtualHelixItemByCoord(*vh.coord())
        activeBaseIdx = partItem.part().activeBaseIndex()
        isActiveNow = vh.hasStrandAtIdx(activeBaseIdx)
        vhi.setActiveSliceView(isActiveNow, activeBaseIdx)
    # end def

    def updateIndexSlot(self, sender, newActiveSliceZIndex):
        part = self.part()
        if part.numberOfVirtualHelices() == 0:
            return
        newlyActiveVHs = set()
        activeBaseIdx = part.activeBaseIndex()
        for vhi in self._partItem._virtualHelixHash.values():
            vh = vhi.virtualHelix()
            if vh:
                isActiveNow = vh.hasStrandAtIdx(activeBaseIdx)
                vhi.setActiveSliceView(isActiveNow, activeBaseIdx)
    # end def

    def updateRectSlot(self, part):
        pass
    # end def

    ### ACCESSORS ###
    def part(self):
        return self._partItem.part()
    # end def

    ### PUBLIC METHODS FOR DRAWING / LAYOUT ###
    def removed(self):
        self._partItem = None
        self._controller.disconnectSignals()
        self.controller = None
    # end def



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
solidrootitem.py
Created by Simon Breslav on 2011-10-05.
"""

from cadnano2.controllers.viewrootcontroller import ViewRootController
from .partitem import PartItem
import maya.cmds as cmds

import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QGraphicsRectItem'])


class SolidRootItem(object):
    """
    SolidRootItem is the root item in the SolidView, it does not have a
    visual representation in the 3D view
    """
    def __init__(self, parent, document, modState):
        self._document = document
        self._controller = ViewRootController(self, document)
        self._modelPart = None
        self._partItems = {}
        self._modifyState = modState
        self._selectionFilterDict = {}
        self.oldSelectionStrandList = []
        self._selectedDict = {}

    ### SLOTS ###
    def partAddedSlot(self, sender, modelPart):
        """
        Receives notification from the model that a part has been added.
        """
        #print "SolidRootItem.partAddedSlot!"
        self._modelPart = modelPart
        partItem = PartItem(modelPart, self)
        self._partItems[partItem] = True
        self.setModifyState(self._modifyState)

    def selectedChangedSlot(self, itemDict):
        """docstring for selectedChangedSlot"""
        pass
    # end def

    def selectionFilterChangedSlot(self, filterNameList):
        self.clearSelectionFilterDict()
        for filterName in filterNameList:
            self.addToSelectionFilterDict(filterName)
    # end def

    def resetRootItemSlot(self, doc):
        pass
    # end def

    def clearSelectionsSlot(self, doc):
        cmds.select(clear=True)
        for strand in list(self._selectedDict.keys()):
            self.removeFromSelectionDict(strand)
    # end def
    ### METHODS ###

    def document(self):
        return self._document
    # end def

    def selectedChanged(self, strandList, value):
        """is called from Mom when selection changed in 3D"""
        # print "selected changed 3D", strandList, self._selectedDict
        if len(strandList) == 0 and len(self._selectedDict) == 0:
            return
        for strand in list(self._selectedDict.keys()):
            self.removeFromSelectionDict(strand)
            temp = self._document.removeStrandFromSelection(strand)
            # print temp, "cupcake"
        self._document.updateSelection()
        for strand in strandList:
            self.addToSelectionDict(strand)
            self._document.addStrandToSelection(strand, value)

        self._document.updateSelection()
        # self.oldSelectionStrandList = strandList
        # pass
    # end def

    def addToSelectionDict(self, item):
        self._selectedDict[item] = True
    # end def

    def removeFromSelectionDict(self, item):
        if item in self._selectedDict:
            del self._selectedDict[item]
    # end def

    def partItems(self):
        """Return a list of partItems associated with this RootItem"""
        return self._partItems
    # end def

    def removePartItem(self, partItem):
        """Remove a given partItem"""
        del self._partItems[partItem]
    # end def

    def setModifyState(self, val):
        """Notifies all the associated partItems of the Modify State Change"""
        self._modifyState = val
        for p in self._partItems:
            p.setModifyState(val)
    # end def

    def clearSelectionFilterDict(self):
        self._selectionFilterDict = {}
    # end def

    def selectionFilterDict(self):
        return self._selectionFilterDict
    # end def

    def addToSelectionFilterDict(self, filterName):
        self._selectionFilterDict[filterName] = True
    # end def

    def removeFromSelectionFilterDict(self, filterName):
        del self._selectionFilterDict[filterName]
    # end def
# end class



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
stranditem.py
Created by Simon Breslav on 2011-10-05.
"""

from controllers.mayacontrollers.mayaObjectManager import Mom
from cadnano2.controllers.itemcontrollers.strand.stranditemcontroller \
                                                import StrandItemController
from cadnano2.model.enum import StrandType
from cadnano2.model.enum import LatticeType

from cadnano2.cadnano import app
import maya.OpenMayaUI as mui
import maya.OpenMaya as mo
import maya.cmds as cmds
import cadnano2.util as util

util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'pyqtSlot', \
                                        'QObject', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QColor'])


class StrandItem(object):
    """
    StrandItem is the visual representation of the strand in the 3D SolidView.
    For this visual representation, StrandItem creates HalfCylinderHelixNode
    Node inside of Maya, so while the StrandItem itself does not get drawn in
    any way, it is the object that communicates with Maya Nodes associated
    with a given strand.
    """
    def __init__(self, mID, modelStrand, virtualHelixItem):
        """
        The parent should be a VirtualHelixItem.
        Initialize function creates the Maya Node for the strand, and setups
        the lookup tables inside of mayaObjectManager (Mom) so that the Maya
        Node can be globally found given a strand, and the other way around.
        Also, sets up StrandItemController that is used to setup all the
        slots and signals between strand model and this strandItem.
        """
        self._modelStrand = modelStrand
        self._virtualHelixItem = virtualHelixItem
        self._viewroot = app().activeDocument.win.solidroot
        mayaNodeInfo = ()
        # print "solidview.StrandItem.__init__ %s" % mID
        if(modelStrand.strandSet().isScaffold()):
            mayaNodeInfo = self.createMayaHelixNodes(virtualHelixItem.x(),
                                                     virtualHelixItem.y(),
                                      modelStrand.oligo().color(),
                                      StrandType.Scaffold,
                                      mID)
        else:
            mayaNodeInfo = self.createMayaHelixNodes(virtualHelixItem.x(),
                                                     virtualHelixItem.y(),
                                      modelStrand.oligo().color(),
                                      StrandType.Staple,
                                      mID)
        #self.onStrandDidMove(strand)
        m = Mom()
        m.cnToMaya[modelStrand] = mayaNodeInfo
        m.mayaToCn[mayaNodeInfo[2]] = modelStrand
        m.mayaToCn[mayaNodeInfo[0]] = modelStrand
        self.updateSize()
        self._controller = StrandItemController(self, modelStrand)
    # end def

    ### SLOTS ###
    def strandResizedSlot(self, strand, indices):
        """Receives notification from the model when a strand is resized"""
        #print "solid.StrandItem.strandResizedSlot", self._modelStrand.idxs()
        self.updateSize()
        self._virtualHelixItem.updateDecorators()
        m = Mom()
        m.updateSelectionBoxes()

    def strandUpdateSlot(self, strand):
        """strandUpdateSlot - empty"""
        pass

    def sequenceAddedSlot(self, oligo):
        """sequenceAddedSlot - empty"""
        pass

    def sequenceClearedSlot(self, oligo):
        """sequenceClearedSlot - empty"""
        pass

    def strandRemovedSlot(self, strand):
        """
        Receives notification from the model when a strand is removed.
        Deletes the strand related mapping in mayaObjectManager, deletes all
        the Maya nodes, deletes all the decorators(live in the
        virtualHelixItem right now), deletes itself from the virtualHelixItem,
        and disconnects itself from the controller.
        """
        mom = Mom()
        mID = mom.strandMayaID(strand)
        mom.removeIDMapping(mID, strand)
        # print "solidview.StrandItem.strandRemovedSlot %s" % mID
        transformName = "%s%s" % (mom.helixTransformName, mID)
        cylinderName = "%s%s" % (mom.helixNodeName, mID)
        meshName = "%s%s" % (mom.helixMeshName, mID)
        if cmds.objExists(transformName):
            cmds.delete(transformName)
        if cmds.objExists(cylinderName):
            cmds.delete(cylinderName)
        if cmds.objExists(meshName):
            cmds.delete(meshName)
        if mID in self._virtualHelixItem.StrandIDs():
            self._virtualHelixItem.StrandIDs().remove(mID)
        self._virtualHelixItem.updateDecorators()
        self._virtualHelixItem.removeStrandItem(self)
        self._virtualHelixItem = None
        self._modelStrand = None
        self._controller.disconnectSignals()
        self._controller = None
    # end def

    def oligoAppearanceChangedSlot(self, oligo):
        """
        Receives notification from the model when a oligo changes appearance.
        Updates the color of the strandItem associated with this strand
        """
        mom = Mom()
        id = mom.strandMayaID(self._modelStrand)
        self.updateColor(id, oligo.color())
        pass

    def oligoSequenceAddedSlot(self, oligo):
        """oligoSequenceAddedSlot - empty"""
        pass

    def oligoSequenceClearedSlot(self, oligo):
        """oligoSequenceClearedSlot - empty"""
        pass

    def strandHasNewOligoSlot(self, strand):
        """
        Receives notification from the model when there is a new oligo.
        Updates the color of the strandItem associated with this strand
        """
        mom = Mom()
        self._controller.reconnectOligoSignals()
        mID = mom.strandMayaID(strand)
        self.updateColor(mID, strand.oligo().color())

    def strandInsertionAddedSlot(self, strand, insertion):
        """strandInsertionAddedSlot - empty"""
        pass

    def strandInsertionChangedSlot(self, strand, insertion):
        """strandInsertionChangedSlot - empty"""
        pass

    def strandInsertionRemovedSlot(self, strand, index):
        """strandInsertionRemovedSlot - empty"""
        pass

    def strandDecoratorAddedSlot(self, strand, decorator):
        """strandDecoratorAddedSlot - empty"""
        pass

    def strandDecoratorChangedSlot(self, strand, decorator):
        """strandDecoratorChangedSlot - empty"""
        pass

    def strandDecoratorRemovedSlot(self, strand, index):
        """strandDecoratorRemovedSlot - empty"""
        pass

    def strandModifierAddedSlot(self, strand, modifier):
        """strandModifierAddedSlot - empty"""
        pass

    def strandModifierChangedSlot(self, strand, modifier):
        """strandModifierChangedSlot - empty"""
        pass

    def strandModifierRemovedSlot(self, strand, index):
        """strandModifierRemovedSlot - empty"""
        pass

    def selectedChangedSlot(self, strand, indices):
        #print "solidview.stranditem.selectedChangedSlot", strand, indices

        mom = Mom()
        if mom.ignoreExternalSelectionSignal:
            return
        mID = mom.strandMayaID(strand)
        mom.ignoreExternalSelectionSignal = True
        transformName = "%s%s" % (mom.helixTransformName, mID)
        if cmds.objExists(transformName):
            if(indices[0] or indices[1]):
                cmds.select(transformName, add=True)
                # print "selecting a strand"
                self._viewroot.addToSelectionDict(strand)
            else:
                # print "deselecting in the slot"
                cmds.select(transformName, deselect=True)
                self._viewroot.removeFromSelectionDict(strand)
        mom.ignoreExternalSelectionSignal = False
    # end def

    ### METHODS ###
    def createMayaHelixNodes(self, x, y, colorname, strandType, mID):
        """
        Create all the Maya nodes, set the initial attributes and connections.
        There are 3 Maya nodes associated with each Strand: Transform Node,
        Shape Node (spHalfCylinderHelixNode), and a Mesh Node (a generic Maya
        Node that is used for rendering) The Mesh Node is the child of the
        Transform Node, and spHalfCylinderHelixNode node inputs the shape data
        into the Mesh Node, using cmds.connectAttr command
         ________________
        | Transform Node |
         ----------------
                |
         ________________  .inMesh          .outputMesh _____________________
        |   Mesh Node    |<---------------------------| HalfCylinderHelixNode |
         ----------------                              -----------------------
        """
        m = Mom()
        cylinderName = "%s%s" % (m.helixNodeName, mID)
        transformName = "%s%s" % (m.helixTransformName, mID)
        meshName = "%s%s" % (m.helixMeshName, mID)
        # shaderName = "%s%s" % (m.helixShaderName, mID)
        cmds.createNode("transform", name=transformName, skipSelect=True)
        cmds.setAttr("%s.rotateX" % transformName, 90)
        cmds.setAttr("%s.translateX" % transformName, x)
        cmds.setAttr("%s.translateY" % transformName, y)
        cmds.createNode("mesh",
                        name=meshName,
                        parent=transformName,
                        skipSelect=True)
        cmds.createNode("spHalfCylinderHelixNode",
                        name=cylinderName,
                        skipSelect=True)
        cmds.connectAttr("%s.outputMesh" % cylinderName,
                         "%s.inMesh" % meshName)
        # XXX - [SB] This should go away and we will ask the model for
        # the right numbers...
        vhi = self._virtualHelixItem
        part = vhi.partItem().part()
        cSType = part.crossSectionType()
        cmds.setAttr("%s.rotation" % cylinderName, part.twistPerBase())
        cmds.setAttr("%s.parity" % cylinderName, vhi.isEvenParity())
        if cSType == LatticeType.Honeycomb:
            cmds.setAttr("%s.rotationOffset" % cylinderName, 250)
            cmds.setAttr("%s.decoratorRotOffset" % cylinderName, 90)
        elif cSType == LatticeType.Square:
            cmds.setAttr("%s.rotationOffset" % cylinderName, 125)
            cmds.setAttr("%s.decoratorRotOffset" % cylinderName, 200)
        else:
            raise NotImplementedError
        cmds.setAttr("%s.strandType" % cylinderName, strandType)
        self.updateColor(mID, colorname)

        cmds.select(transformName)
        cmds.polySoftEdge(a=89.99)
        cmds.setAttr("%s.displayEdges" % meshName, 2)
        cmds.select(clear=True)
        return (cylinderName, transformName, meshName)

    def updateColor(self, mID, colorname):
        """
        Update the color of the Maya's Mesh Node associated with a this
        StrandItem, this is done by creating a shadingNode for each color or
        connecting the Mesh Mode to an existing shadingNode if one exists
        for a given color.
        """
        m = Mom()
        meshName = "%s%s" % (m.helixMeshName, mID)
        color = QColor(colorname)
        colorval = "%d_%d_%d" % (color.red(), color.green(), color.blue())
        shaderName = "%s%d_%d_%d" % (m.helixShaderName, color.red(),
                                                  color.green(),
                                                  color.blue())
        if not cmds.objExists(shaderName):
            # Shader does not exist create one
            cmds.shadingNode('lambert', asShader=True, name=shaderName)
            cmds.sets(n="%sSG" % shaderName, r=True, nss=True, em=True)
            cmds.connectAttr("%s.outColor" % shaderName,
                             "%sSG.surfaceShader" % shaderName)
            cmds.setAttr("%s.color" % shaderName,
                         color.redF(), color.greenF(), color.blueF(),
                         type="double3")
            cmds.sets(meshName, forceElement="%sSG" % shaderName)
        else:
            #shader exist connect
            cmds.sets(meshName, forceElement="%sSG" % shaderName)

    def updateSize(self):
        """
        Update Maya's Half Cylinder Node attributes related to the size
        """
        mom = Mom()
        mID = mom.strandMayaID(self._modelStrand)
        cylinderName = "%s%s" % (mom.helixNodeName, mID)
        endpoints = self._modelStrand.idxs()
        totalNumBases = \
                self._virtualHelixItem.virtualHelix().part().maxBaseIdx()
        cmds.setAttr("%s.startBase" % cylinderName,
                             endpoints[0])

        cmds.setAttr("%s.endBase" % cylinderName,
                             endpoints[1])
        cmds.setAttr("%s.totalBases" % cylinderName, int(totalNumBases))



###############################################################################
#
# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
###############################################################################


import os
import sys
import maya
import maya.OpenMaya as OpenMaya
import maya.OpenMayaMPx as OpenMayaMPx
import maya.cmds as cmds
import maya.OpenMayaUI as OpenMayaUI
import maya.mel as mel
import sip

sys.path.insert(0, os.environ['CADNANO_PATH'])
# cmds.flushUndo()
# cmds.undoInfo(state=False)

import cadnano2.views.solidview.mayaHotKeys as mayaHotKeys
import cadnano2.views.solidview.mayaUI as mayaUI

import cadnano2.util as util
util.qtFrameworkList = ['PyQt'] # necessary to overide defaults
util.qtWrapImport('QtGui', globals(), ['qApp', 'QDockWidget', 'QSizePolicy'])

util.qtWrapImport('QtCore', globals(), ['Qt', 'QObject'])

kPluginName = "spCadNano"
gCadNanoButton = None
gCadNanoToolbar = None
fMayaExitingCB = None

gCadNanoApp = None

gIconPath = (
        os.environ['CADNANO_PATH'] +
        "/ui/mainwindow/images/cadnano2-app-icon_shelf.png")


# command
class openCadNano(OpenMayaMPx.MPxCommand):
    def __init__(self):
        OpenMayaMPx.MPxCommand.__init__(self)

    def doIt(self, argList):
        openCN()

    @staticmethod
    def creator():
        return OpenMayaMPx.asMPxPtr(openCadNano())


class closeCadNano(OpenMayaMPx.MPxCommand):
    def __init__(self):
        OpenMayaMPx.MPxCommand.__init__(self)

    def doIt(self, argList):
        closeCN()

    @staticmethod
    def creator():
        return OpenMayaMPx.asMPxPtr(closeCadNano())


def onExitingMaya(clientData):
    closeCN()
    mel.eval("savePrefs;")


def onHideEvent():
    closeCN()


# Initialize the script plug-in
def initializePlugin(mobject):
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.registerCommand("openCadNano", openCadNano.creator)
        mplugin.registerCommand("closeCadNano", closeCadNano.creator)
    except:
        sys.stderr.write("Failed to register command: %s\n" % kPluginName)
        raise
    addUIButton()
    global fMayaExitingCB
    fMayaExitingCB = OpenMaya.MSceneMessage.addCallback(
                                OpenMaya.MSceneMessage.kMayaExiting,
                                onExitingMaya)


# Uninitialize the script plug-in
def uninitializePlugin(mobject):
    global gCadNanoApp
    closeCN()
    removeUIButton()

    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.deregisterCommand("openCadNano")
        mplugin.deregisterCommand("closeCadNano")
    except:
        sys.stderr.write("Failed to unregister command: %s\n" % kPluginName)
        raise

    global fMayaExitingCB
    if (fMayaExitingCB != None):
        OpenMaya.MSceneMessage.removeCallback(fMayaExitingCB)


def openCN():
    global gCadNanoApp
    modifyMayaUI()
    isInit = False
    if gCadNanoApp:
        for x in gCadNanoApp.documentControllers:
            if x.win:
                x.windock.setVisible(True)
                x.win.setVisible(True)
    else:
        # begin program
        import cadnano
        gCadNanoApp =  cadnano.initAppMaya()
        isInit = True
        
    if gCadNanoApp.activeDocument:
        if hasattr(gCadNanoApp.activeDocument, 'solidHelixGrp'):
            if gCadNanoApp.activeDocument.solidHelixGrp:
                gCadNanoApp.activeDocument.solidHelixGrp.onPersistentDataChanged()

    pluginPath = os.path.join(os.environ['CADNANO_PATH'],
                                "controllers", "mayacontrollers")
    hmPath = os.path.join(pluginPath, "helixManip.py")
    msPath = os.path.join(pluginPath, "mayaSelectionContex.py")
    rmPath = os.path.join(pluginPath, "removedMsgCmd.py")
    if not cmds.pluginInfo(rmPath, query=True, loaded=True):
        cmds.loadPlugin(rmPath)
        cmds.spRemovedMsg()
    if not cmds.pluginInfo(msPath, query=True, loaded=True):
        cmds.loadPlugin(msPath)
        cmds.spMayaCtxCmd("spMayaCtxCmd1")
        cmds.setToolTo("spMayaCtxCmd1")
    if not cmds.pluginInfo(hmPath, query=True, loaded=True):
            cmds.loadPlugin(hmPath)
    
    if isInit:
        gCadNanoApp.finishInit()
# end def

def changed(self, event):
    print(str(event.type()))
    if (event.type() == QEvent.ActivationChange or
        event.type() == QEvent.WindowActivate or
        event.type() == QEvent.ApplicationActivate):
            print(self.win.windowTitle())
            app().activeDocument = self


def modifyMayaUI():
    mayaHotKeys.disableAllHotKeys()
    mayaUI.simplifyUI()
    mayaUI.setViewportQuality()

    myWindow = cmds.window()
    myForm = cmds.formLayout(parent=myWindow)
    global gCadNanoToolbar
    gCadNanoToolbar = cmds.toolBar(
                                "cadnanoBox",
                                area='top',
                                allowedArea='top',
                                content=myWindow)

    global gIconPath
    closeCadNanoCmd = 'import maya.cmds;maya.cmds.closeCadNano()'
    myButton = cmds.iconTextButton(
                               label='Quit cadnano',
                               annotation='Quit cadnano interface',
                               image1=gIconPath,
                               parent=myForm,
                               command=closeCadNanoCmd)
    cmds.formLayout(
                myForm,
                edit=True,
                attachForm=[(myButton, 'right', 10)])


def restoreMayaUI():
    mayaHotKeys.restoreAllHotKeys()
    mayaUI.restoreUI()
    mayaUI.restoreViewportQuality()

    if gCadNanoToolbar:
        if cmds.toolBar(gCadNanoToolbar, exists=True):
            cmds.deleteUI(gCadNanoToolbar)


def closeCN():
    global gCadNanoApp
    if gCadNanoApp:
        if gCadNanoApp.activeDocument:
            gCadNanoApp.activeDocument.win.setVisible(False)
            gCadNanoApp.activeDocument.windock.setVisible(False)
        for x in gCadNanoApp.documentControllers:
            if x.win:
                x.windock.setVisible(False)
                x.win.setVisible(False)
    restoreMayaUI()


def addUIButton():
    global gCadNanoButton
    global gIconPath
    mayaMainToolbar = (
            'MayaWindow|toolBar1|MainStatusLineLayout|formLayout5|formLayout8')
    if cmds.formLayout(mayaMainToolbar, ex=True):
        cmds.setParent(mayaMainToolbar)
        gCadNanoButton = cmds.iconTextButton(
                         label='cadnano',
                         annotation='Launch cadnano interface',
                         image1=gIconPath,
                         parent=mayaMainToolbar,
                         command='import maya.cmds; maya.cmds.openCadNano()')
        cmds.formLayout(
                    mayaMainToolbar,
                    edit=True,
                    attachForm=[(gCadNanoButton, 'right', 10)])


def removeUIButton():
    global gCadNanoButton
    cmds.deleteUI(gCadNanoButton)
